import os, re
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

try:
    import steammaterials  # TESTING PURPOSES

    steammaterials_flag = False
except:
    steammaterials_flag = True

from scipy import constants, optimize, integrate, signal, interpolate
from fiqus.data.DataFiQuSConductorAC_Strand import CACStrandGeometry, CACStrandMesh, CACStrandSolve
from fiqus.plotters.PlotPythonConductorAC import BatchPlotPythonConductorAC, PlotPythonConductorAC, PDFreport
from fiqus.utils.Utils import FilesAndFolders
import ruamel.yaml


class SimulationData:
    """
    Class used to store and manage data from a single simulation.

    This class is responsible for loading and organizing the data from a single simulation.
    It stores the data in various attributes and provides methods for retrieving and processing the data.

    """

    def __init__(self, model_data_output_path, geometry_name, mesh_name, solution_name, fdm) -> None:
        self.model_data_output_path = model_data_output_path  # This is the path to the folder where the model output data is stored (e.g. geometries)
        self.geometry_name = geometry_name  # Name of the geometry folder
        self.mesh_name = mesh_name  # Name of the mesh folder
        self.solution_name = solution_name  # Name of the solution folder
        self.conductor_name = fdm.magnet.solve.conductor_name

        # Organize the folders:
        self.geometry_folder = os.path.join(self.model_data_output_path, geometry_name)  # Path to the geometry folder
        self.mesh_folder = os.path.join(self.geometry_folder, mesh_name)  # Path to the mesh folder
        self.solution_folder = os.path.join(self.mesh_folder, solution_name)  # Path to the solution folder

        # hacky solution for now ... pass conductor data through fdm because its not stored in geom yaml
        self.conductor = fdm.conductors[self.conductor_name]
        # Store the YAML input-files in a data model, fdm:
        self.geometry, self.mesh, self.solve = self.retrieve_fiqusDataModel()
        # Store losses, simulation time and check if the simulation crashed:
        temp_file_path = os.path.join(self.solution_folder, 'text_output')
        loss_file = [f for f in os.listdir(temp_file_path) if f.startswith('power') and f.endswith('.txt')][0]
        self.power_columns = ['Time', 'FilamentLoss', 'CouplingLoss', 'EddyLoss', 'TotalLoss', 'CouplingLoss_dyn', 'TotalLoss_dyn']  # Only in the case dynamic correction is used, must be changed later
        self.power = pd.read_csv(os.path.join(self.solution_folder, 'text_output', loss_file), sep=' ', names=self.power_columns)  # Store instantaneous losses as pandas dataframe
        # Add a row of zeros at the beginning of the dataframe to account for the initial condition:
        self.power = pd.concat([pd.DataFrame({col: 0 for col in self.power_columns}, index=[0]), self.power]).reset_index(drop=True)
        if self.solve.formulation_parameters.formulation == "CATI" and self.solve.diffusion_barriers:  # Check if dm.solve has the entry for diffusion barriers
            if self.solve.diffusion_barriers.enable:  # Check if diffusion barriers are enabled
                # Load and add the power dissipated in the diffusion barriers (if they are enabled):
                self.power_diffusion_barriers = np.loadtxt(os.path.join(temp_file_path, 'power_diffusion_barrier.txt'), skiprows=1)
                self.power_diffusion_barriers = np.sum(self.power_diffusion_barriers[:, 1:], axis=1)
                self.power['CouplingLoss'] += self.power_diffusion_barriers
                self.power['CouplingLoss_dyn'] += self.power_diffusion_barriers
                self.power['TotalLoss'] += self.power_diffusion_barriers
                self.power['TotalLoss_dyn'] += self.power_diffusion_barriers
        self.crash = True if 'crash_report.txt' in os.listdir(temp_file_path) else False
        # Store simulation time:
        try:
            with open(os.path.join(self.solution_folder, 'text_output', 'simulation_time.txt'), 'r') as f:
                self.simulation_time = float(f.readline().strip())
        except:
            self.simulation_time = None  # If the simulation time file does not exist, the simulation has not finished running.

        # Store the rest of the post-processing data:
        self.time = self.power['Time']
        self.instantaneous_temperature = self.load_standard_data(os.path.join(temp_file_path, 'temperature.txt'), 1, add_initial_zero=True)
        self.temperature = self.load_standard_data(os.path.join(temp_file_path, 'temperature.txt'), 2, add_initial_zero=True)
        self.I_transport = self.load_standard_data(os.path.join(temp_file_path, 'I_transport.txt'), 1)
        self.V_transport = self.load_standard_data(os.path.join(temp_file_path, 'V_transport.txt'), 1)
        self.V_transport_unitlen = self.load_standard_data(os.path.join(temp_file_path, 'V_transport_unitlen.txt'), 1)
        self.hs_val = self.load_standard_data(os.path.join(temp_file_path, 'hs_val.txt'), 1)
        self.magn_fil = self.load_standard_data(os.path.join(temp_file_path, 'magn_fil.txt'), [1, 2, 3])
        self.magn_matrix = self.load_standard_data(os.path.join(temp_file_path, 'magn_matrix.txt'), [1, 2, 3])
        self.I = self.load_standard_data(os.path.join(temp_file_path, 'I.txt'), 1, len(self.time))
        self.V = self.load_standard_data(os.path.join(temp_file_path, 'V.txt'), 1, len(self.time))
        self.I_integral = self.load_standard_data(os.path.join(temp_file_path, 'I_integral.txt'), 1, len(self.time))
        self.I_abs_integral = self.load_standard_data(os.path.join(temp_file_path, 'I_abs_integral.txt'), 1, len(self.time))
        self.magnetic_energy_internal = self.load_standard_data(os.path.join(temp_file_path, 'magnetic_energy_internal.txt'), 1)
        self.magnetic_energy_external = self.load_standard_data(os.path.join(temp_file_path, 'magnetic_energy_external.txt'), 1)
        self.flux_external = self.load_standard_data(os.path.join(temp_file_path, 'flux_external.txt'), 1)
        self.flux_internal = self.load_standard_data(os.path.join(temp_file_path, 'flux_internal.txt'), 1)
        self.selffield_dfluxdt = self.load_standard_data(os.path.join(temp_file_path, 'selffield_dfluxdt.txt'), 1)
        self.Ip = self.load_standard_data(os.path.join(temp_file_path, 'Ip.txt'), 1, len(self.time))
        self.Vp = self.load_standard_data(os.path.join(temp_file_path, 'Vp.txt'), 1, len(self.time))
        # try to init - otherwise NONE
        self.f, self.B_background, self.I_amp, self.T_background = self.get_source_params(verbose=True)
        # self.data_impedance = self.compute_data_impedance(verbose=False)
        # self.analytic_RL = self.compute_analytic_RL()
        # Integrate the losses to obtain the cumulative power and the total power per cycle:
        self.cumulative_power, self.total_power_per_cycle = self.integrate_power()  # Store cumulative power and total cumulative power per cycle

    def export2txt(self, data_names, data_columns, filename):
        """ This function can be used to export one or more arrays of simulation data in one txt file (stored in solution folder)."""
        np_data = np.array(data_columns).T
        np.savetxt(os.path.join(self.solution_folder, str(filename) + '.txt'), np_data, header=" ".join(data_names), comments="")

    def rad_to_deg(self, radiants):
        return radiants / np.pi * 180

    def get_source_params(self, verbose=False):
        source = self.solve.source_parameters.source_type
        T = self.solve.general_parameters.temperature
        if source == "rotating":
            f = self.solve.source_parameters.rotating.frequency
            B = self.solve.source_parameters.rotating.field_magnitude
            I = 0
        elif source == "sine":
            f = self.solve.source_parameters.sine.frequency
            I = self.solve.source_parameters.sine.current_amplitude
            # background field amplitude
            B = self.solve.source_parameters.sine.superimposed_DC.field_magnitude
            if verbose: print("Loading Simulation: f=" + "{:.2f}".format(f) + " [Hz];  B_back=" + "{:.2f}".format(B) + " [T];  I=" + "{:.2f}".format(I) + " [A]")
        else:
            f = None
            B = 0
            I = max(abs(self.I_transport))

        return f, B, I, T

    def first_rise(self, current=False):
        """ returns the time, current and flux on the first penetration interval until the first current peak/saddle is reached """
        first_extrema_idx = signal.argrelextrema(self.I_transport if current else self.flux_internal, np.greater_equal)[0][0]  # first rel extrema
        return self.time[0:first_extrema_idx], self.I_transport[0:first_extrema_idx], self.flux_internal[0:first_extrema_idx]

    def halfcycle(self, current=False):
        """ returns the time, current and flux on the first halfcycle of transport current or flux """
        data = self.I_transport if current else self.flux_internal
        zero_crossings = np.where(np.diff(np.sign(data[1:])))[0] + 1
        return self.time[0:zero_crossings[0]], self.I_transport[0:zero_crossings[0]], self.flux_internal[0:zero_crossings[0]]

    def load_standard_data(self, file_path, columns, reshape=None, add_initial_zero=False):
        """
        There are many output .txt-files with similar format. This function loads the data from one of these files and returns it as a numpy array.
        If the file does not exist, None is returned without raising an error.
        """
        try:
            data = np.loadtxt(file_path, comments='#', usecols=columns)
            if reshape:
                data = data.reshape(-1, reshape).T
        except IOError:
            return None

        if add_initial_zero:
            if len(data.shape) == 1:
                data = np.insert(data, 0, 0)
            else:
                zeros = np.zeros((1, data.shape[1]))
                data = np.vstack((zeros, data))
        return data

    def retrieve_fiqusDataModel(self):
        """
        This function reads the YAML input-files for geometry, mesh and solve and stores them in three dictionaries which are returned.
        This function is to be called only once, when the object is created.
        """
        geometry_dataModel = FilesAndFolders.read_data_from_yaml(os.path.join(self.geometry_folder, 'geometry.yaml'), CACStrandGeometry)
        mesh_dataModel = FilesAndFolders.read_data_from_yaml(os.path.join(self.mesh_folder, 'mesh.yaml'), CACStrandMesh)
        # Now that the solve.yaml file contains more than just the solve part, we query only that part here.
        # solution_dataModel = FilesAndFolders.read_data_from_yaml(os.path.join(self.solution_folder, 'solve.yaml'), CACStrandSolve)
        with open(os.path.join(self.solution_folder, 'solve.yaml'), 'r') as stream:
            yaml = ruamel.yaml.YAML(typ='safe', pure=True)
            yaml_str = yaml.load(stream)
            solution_dataModel = CACStrandSolve(**yaml_str['solve'])
        return geometry_dataModel, mesh_dataModel, solution_dataModel

    def integrate_power(self):
        """
        This function integrates the instantaneous power over time to obtain the cumulative power.
        It also calculates the total cumulative power per cycle.
        The cumulative power is returned as a pandas dataframe and the total cumulative power per cycle is returned as a dictionary.
        """
        find_closest_idx = lambda arr, val: np.abs(arr - val).argmin()

        t = np.array(self.power['Time'])
        t_final = t[-1]
        if self.f:
            t_init = find_closest_idx(t, t_final - 1 / (self.f))
        else:
            t_init = find_closest_idx(t, t_final)

        cumulative_power = pd.DataFrame(columns=self.power_columns)
        total_power_per_cycle = {}

        cumulative_power['Time'] = self.power["Time"]
        for column in self.power_columns[1:]:
            cumulative_power[column] = np.insert(integrate.cumulative_trapezoid(self.power[column], t), 0, 0)
            total_power_per_cycle[column] = (cumulative_power[column].iloc[-1] - cumulative_power[column].iloc[t_init])  # / (np.pi*matrix_radius**2 * loss_factor) # Why do we divide by pi*matrix_radius**2*loss_factor?

        return cumulative_power, total_power_per_cycle

    def compute_data_impedance(self, verbose=False):
        """
        This function estimates a complex impedance per unit length for the OOP based on sinus fits for the transport voltage and current.
        If no sinus fit is possible it returns NONE as impedance estimate.
        """
        if not self.f:
            return None

        I_mean = np.mean(self.I_transport)
        V_mean = np.mean(self.V_transport_unitlen)
        I_peak = max(abs(self.I_transport))
        V_peak = max(abs(self.V_transport_unitlen))
        # ignore values before this index - tweaking parameter depending on source ramping
        N_h = round(len(self.I_transport) / 4)
        # sine fit for current and voltage
        optimize_func_current = lambda x: x[0] * np.sin(x[1] * self.time[N_h:] + x[2]) + x[3] - self.I_transport[N_h:]
        optimize_func_voltage = lambda x: x[0] * np.sin(x[1] * self.time[N_h:] + x[2]) + x[3] - self.V_transport_unitlen[N_h:]
        current_estimate = optimize.leastsq(optimize_func_current, [I_peak - I_mean, self.f, 0, I_mean])[0]  # est_amp, est_freq, est_phase, est_mean
        voltage_estimate = optimize.leastsq(optimize_func_voltage, [V_peak - V_mean, self.f, 0, V_mean])[0]
        delta_phi = abs(voltage_estimate[2] - current_estimate[2]) % (2 * np.pi)
        Z_magnitude = V_peak / I_peak  # both sinus - no formfactor
        L_inductance = abs(np.sin(delta_phi) * Z_magnitude) / (2 * np.pi * self.f)
        R_resistance = abs(np.cos(delta_phi) * Z_magnitude)
        Z_impedance = R_resistance + 2 * np.pi * self.f * L_inductance * 1j
        if verbose:
            print("estimated simulation Impedance: " + str(Z_magnitude) + " Ohm")
            print(" -> L: " + str(L_inductance) + " Henry")
            print(" -> R: " + str(R_resistance) + " Ohm")
            print(" -> Phase: " + str(np.angle(Z_impedance, deg=True)) + " Deg")
        return Z_impedance

    def compute_analytic_RL(self):
        """ This function calculates the analytic impedance parameters R & L per unit length based on reduced telegraphers equations with C, G = 0 """

        rho_CU = 1.81e-10
        coil_radius = self.geometry.coil_radius
        air_radius = self.geometry.air_radius
        strand_radius = self.conductor.strand.diameter / 2
        filament_radius = self.conductor.strand.filament_diameter / 2

        correction_factor = 0.9549  # CATI (needs to be corrected)
        ell = correction_factor * self.conductor.strand.fil_twist_pitch / 6  # CATI

        if self.geometry.type == 'strand_only':
            # single filament in open space
            inductance_L = constants.mu_0 / (2 * np.pi) * np.log(air_radius / filament_radius)  # == coax cable
            resistance_R = rho_CU / (np.pi * (self.conductor.strand.diameter / 2) ** 2) * ell if self.solve.general_parameters.superconductor_linear else 1e-12
        elif self.geometry.type == 'coil':
            # single coil winding closed in infinity
            inductance_L = constants.mu_0 / np.pi * np.log((coil_radius * 2 - strand_radius) / strand_radius)  # + ell * constants.mu_0 / (4 * np.pi)
            resistance_R = 2 * rho_CU / (np.pi * (self.conductor.strand.diameter / 2) ** 2) * ell if self.solve.general_parameters.superconductor_linear else 1e-12
        else:
            return None

        return [resistance_R, inductance_L]


class ROHFmodel:
    """ Base class for a Reduced Order Hysteretic Flux model. Defining a chain of rate dependent cells for flux and loss reconstruction in a homogenized strand model."""

    def __init__(self, N=None, taus=None, kappas_bar=None, alphas=None, simulation=None, filepath=None, filename=None, spacing_type=None, label=''):
        # direct parameter initialization
        self.N = N
        self.kappas_bar = kappas_bar
        self.alphas = alphas
        self.taus = taus
        # Ambient sueprcon parameters
        self.Ic_0 = 2978  # [A]
        self.T_0 = 1.9  # [K]
        self.Ec_0 = 1e-4  # [V/m]
        self.n = 60
        # internal variables
        self.L0 = constants.mu_0 / (4 * np.pi)  # lin. self inductance of cylindric wire
        # optional info for batch plots
        self.label = label
        self.simulation = simulation
        # initialization shortcuts
        if filepath and filename:
            self.import_params(filepath=filepath, filename=filename)  # initialize from file
        elif simulation:  # fit params on the fly
            self.fit_data(simulation=simulation, spacing_type=spacing_type, taus=taus, kappas_bar=kappas_bar, alphas=alphas)

    def update(self, N, taus, kappas_bar, alphas):
        self.N = N
        self.taus = taus
        self.kappas_bar = kappas_bar
        self.alphas = alphas

    def params(self):
        return np.array([self.alphas, self.kappas_bar, self.taus]).T

    def export_params(self, filepath, filename, txt_format=False):
        """Exports the reduced order parameters kappa and alpha as txt file.

        :param filepath: absolute path to storage location
        :param filename: name of the file (without .txt ending)
        """
        if txt_format:
            header = "# fitted on " + self.simulation.solution_name + "\nalphas kappas taus" if self.simulation else "# imported params \nalphas kappas taus"
            np.savetxt(os.path.join(filepath, str(filename) + '.txt'), self.params(), header=header, comments="")
        else:
            df = pd.DataFrame(self.params())
            df.to_csv(os.path.join(filepath, str(filename) + '.csv'), header=['alphas', 'kappas', 'taus'], index=None, float_format='{:e}'.format)

    def import_params(self, filepath, filename, txt_format=False):
        """This function reads the reduced order parameters kappa and alphas from a txt file.

        :param filepath: absolute path to storage location
        :param filename: name of the file (without .txt ending)
        """
        if txt_format:
            data = np.loadtxt(os.path.join(filepath, filename + ".txt"), delimiter=" ", skiprows=2).T  # alphas, kappas, taus
        else:
            data = np.array(pd.read_csv(os.path.join(filepath, filename + ".csv"), sep=',', header=1).values).T
            print(data)
        self.simulation = None
        self.name = filename
        self.update(N=len(data[0]), alphas=data[0], kappas_bar=data[1], taus=data[2])

    def ohmic_loss(self, I, simulation=None):
        """ This function calculates the ohmic loss in filaments and matrix of a homogenized conductor.

        :param I: transport current
        :param simulation: simulation data for conductor information, defaults to None
        :return: array of [p_ohm_fil, p_ohm_mat, R_fil, R_mat] for all current steps
        """

        if steammaterials_flag or simulation is None:
            # print("Assuming zero ohmic loss")
            return np.zeros(len(I)), np.zeros(len(I))

            # geometry parameters
        surf_strand = np.pi * (simulation.conductor.strand.diameter / 2) ** 2
        surf_filaments = simulation.conductor.strand.number_of_filaments * np.pi * (simulation.conductor.strand.filament_diameter / 2) ** 2
        surf_matrix = surf_strand - surf_filaments
        # matrix resistivity
        matpath = os.path.join(os.path.dirname(steammaterials.__file__), 'CFUN')
        Tspace = np.vstack((self.T_0, 0, simulation.conductor.strand.RRR))  # T, B, RRR
        cu_space = steammaterials.SteamMaterials("CFUN_rhoCu_NIST_v2", Tspace.shape[0], Tspace.shape[1], matpath)
        rho_matrix = cu_space.evaluate(Tspace)
        R_matrix = rho_matrix / surf_matrix

        # current sharing between matrix (R const) and filaments (power law) https://ieeexplore.ieee.org/document/8970335
        grid = np.vstack((I / (surf_strand), self.Ec_0 * np.ones(len(I)), self.Ic_0 / (surf_filaments) * np.ones(len(I)), self.n * np.ones(len(I)), rho_matrix * np.ones(len(I)), surf_filaments * np.ones(len(I)), surf_matrix * np.ones(len(I))))
        sharing_space = steammaterials.SteamMaterials("CFUN_HTS_CurrentSharing_homogenized_v1", grid.shape[0], grid.shape[1], matpath)
        lambdas = sharing_space.evaluate(grid)
        # current split associated ohmic losses
        I_filaments = lambdas * I
        I_matrix = (np.ones(len(lambdas)) - lambdas) * I
        p_ohm_filaments = self.Ec_0 * ((abs(I_filaments)) / (self.Ic_0)) ** (self.n) * abs(I_filaments)
        p_ohm_matrix = R_matrix * I_matrix ** 2

        return p_ohm_filaments, p_ohm_matrix

    def fit_scaling(self, simulation):
        """ This method fits the internal kappa_scaling parameter of the ROHFmodel to a given magnetic background field magnitude for NbTi filaments.

        :param B_background: magnitude of the magnetic background field in Tesla.
        """
        verbose = False
        # Scaling params
        B_background = simulation.B_background
        T_background = simulation.T_background
        if verbose: print("ROHF scaling: B=" + str(B_background) + " T=" + str(T_background))

        filament_d = simulation.conductor.strand.filament_diameter
        strand_surf = np.pi * (simulation.conductor.strand.diameter / 2) ** 2
        filament_surf = simulation.conductor.strand.number_of_filaments * np.pi * (filament_d / 2) ** 2
        fill_factor = filament_surf / strand_surf
        Jc_0 = self.Ic_0 / filament_surf

        if steammaterials_flag or B_background is None:
            raise ValueError('cant perform field scaling fit. Need steammaterials and background field amplitude.')
        matpath = os.path.join(os.path.dirname(steammaterials.__file__), 'CFUN')

        # Determine ambient field with no background and max current density
        B_space = np.linspace(0, 14, 5000)
        numpy2d_Bspace = np.vstack((self.T_0 * np.ones(B_space.shape), B_space, 1 * np.ones(B_space.shape)))  # T, B, A
        nbti_space = steammaterials.SteamMaterials("CFUN_IcNbTi_v1", numpy2d_Bspace.shape[0], numpy2d_Bspace.shape[1], matpath)
        jc_nbti_space = nbti_space.evaluate(numpy2d_Bspace)
        ambient_idx = np.abs(jc_nbti_space - Jc_0).argmin()
        B_ambient = B_space[ambient_idx]
        if verbose: print(" Ambient field: " + str(B_ambient) + " [T]")

        # determine absolute field value including a scaled down version of the strand ambient field
        B_abs = B_background + B_ambient
        # determine first scaling factor for the magnetic field
        jc_nbti1 = steammaterials.SteamMaterials("CFUN_IcNbTi_v1", 3, 1, matpath)
        jc_Babs = jc_nbti1.evaluate((T_background, B_abs, 1))
        scaling = jc_Babs / Jc_0
        # determine temperature scaling with finer ambient estimate
        if T_background != self.T_0:
            B_abs = float(B_background + scaling * B_ambient)
            jc_nbti2 = steammaterials.SteamMaterials("CFUN_IcNbTi_v1", 3, 1, matpath)
            jc_Babs = jc_nbti2.evaluate((T_background, B_abs, 1))
            scaling = jc_Babs / Jc_0
        if verbose: print(" Field scaling : " + str(scaling))
        return scaling, fill_factor

    def fit_kappas(self, simulation, spacing_type=None, verbose=False):
        """
        This method fits the coercivity parameters 'kappas_bar' according to a 'spacing_type' and given number of cells 'self.N'.
        The kappas are spread on the virgin rise from zero interval of a given strand simulation and can be seen as list of activation levels for the N cells.

        :param simulation: SimulationData object which the ROHF will get fitted to
        :param spacing_type: Enforce a specfic spacing of the kappas (i.e. 'linear'), defaults to None
        :param verbose: print additional information, defaults to False
        :raises ValueError: Unknown options for spacing_type
        :return: kappas_bar ndarray of N kappas
        """
        _, I_interval, flux_interval = simulation.first_rise()
        if spacing_type:
            if spacing_type == "exponential":
                [a, b], res1 = optimize.curve_fit(lambda x1, a, b: a * np.exp(b * x1), I_interval / max(I_interval), flux_interval / max(flux_interval))
                kappa_spacing = max(I_interval) * np.flip(np.exp(b) - np.exp(b * np.linspace(0, 1, num=self.N + 1))) / (np.exp(b) - 1)
            elif spacing_type == "linear":
                kappa_spacing = np.linspace(min(I_interval), max(I_interval), self.N + 1)
            elif spacing_type == "invlog":
                kappa_spacing = np.insert(max(I_interval) * (1.5 - np.geomspace(1, .5, num=self.N + 1)), 0, 0)
            elif spacing_type == "log":
                kappa_spacing = np.insert(np.geomspace(I_interval[1], max(I_interval), self.N), 0, 0)
            else:
                raise ValueError('Unknown kappa spacing for S_chain fit')
            # get closest matching indices in I_interval as kappas
            kappas = I_interval[self.__closest_matches(I_interval, kappa_spacing)]
        else:
            # calculate kappas based on minimization of linear interpolation error
            def max_abs_error(kappas, flux_spline, I_samples):
                # computes maximum absolute error between rohf and reference simulation on sample points
                rohf_flux = np.interp(I_samples, kappas, flux_spline(kappas))
                return max(abs(rohf_flux - flux_spline(I_samples)))

            # interpolate
            I_samples = np.linspace(0, max(I_interval), 500)
            flux_spline = interpolate.CubicSpline(I_interval, flux_interval, extrapolate=True)
            # minimize on interval
            cons = (optimize.LinearConstraint([np.concatenate(([1], np.zeros(self.N)))], lb=0, ub=0),
                    optimize.LinearConstraint([np.concatenate((np.zeros(self.N), [1]))], lb=max(I_interval), ub=max(I_interval)))
            bounds = ((0, max(I_interval)),) * (self.N + 1)
            kappas = optimize.differential_evolution(max_abs_error, args=(flux_spline, I_samples), x0=np.linspace(0, max(I_interval), self.N + 1),
                                                     disp=True, constraints=cons, bounds=bounds, tol=1e-8).x
            # kappas = np.insert(kappas, 1, 0.08*max(I_interval))
            if verbose:
                plt.figure()
                plt.title("kappa distribution")
                plt.plot(I_interval, flux_interval, '-b')
                plt.plot(kappas, flux_spline(kappas), 'gX')
                plt.show()

        return kappas[:-1]

    def fit_alphas(self, simulation, kappas_bar, dual=False, verbose=False):
        """Calculates the cell weights alpha for each kappa by cellwise recursion, fitting the simulation data given through I_interval and flux_interval.

        :param simulation: SimulationData object which the ROHF will get fitted to
        :param kappas_bar: coercivity currents on virgin I_interval which are used to define N multicell ROHF
        :param verbose: display all information, defaults to False
        :return: kappas_bar, alphas. List of updated kappas and their matching weights. Kappas only change in case of a dual fitting strategy (dual=True)
        """
        _, I_interval, flux_interval = simulation.first_rise(current=True)

        flux_spline = interpolate.CubicSpline(I_interval, flux_interval, extrapolate=True)
        kappas = np.concatenate((kappas_bar, [I_interval[-1]]))  # N + 1 kappas with interval boundary

        # classic coercivity-weight linkage (undershoot PCloss)
        alphas = np.zeros(len(kappas_bar))
        delta_flux = np.diff(flux_spline(kappas))
        delta_current = np.diff(kappas)
        for k in range(0, self.N):
            alphas[k] = 0 if delta_current[k] == 0 else abs(delta_flux[k] / (self.L0 * delta_current[k]) - sum(alphas))
        if verbose: print(" optimized alphas:    " + str(alphas))

        if False:
            # tangential coercivity-weight linkage (overshoot PCloss)
            dflux_dI_spline = flux_spline.derivative(nu=1)
            fluxes = flux_spline(kappas)
            m_slopes = dflux_dI_spline(kappas)
            alphas_tan = np.diff(m_slopes) / self.L0
            m_slopes[0] = 0
            b_offsets = np.zeros(self.N + 1)
            kappas_tan = np.zeros(self.N + 1)
            for k in range(1, self.N + 1):
                b_offsets[k] = fluxes[k] - m_slopes[k] * kappas[k]
                kappas_tan[k] = (b_offsets[k] - b_offsets[k - 1]) / (m_slopes[k - 1] - m_slopes[k])
            kappas_bar_tan = kappas_tan[1:]
            if True:
                plt.figure()
                plt.plot(I_interval, flux_interval, label='reference')
                plt.scatter(kappas_bar, fluxes[:-1], label='k-a-linkage kappas')
                plt.scatter(kappas_bar_tan, np.zeros(len(kappas_bar_tan)), label='new kappas', marker='x')
                for k in range(0, self.N + 1):
                    plt.plot(I_interval, b_offsets[k] + m_slopes[k] * I_interval)
                plt.ylim([0, max(flux_interval)])
                plt.legend()
                plt.plot()
                plt.show()

        return kappas_bar, 0.97 * alphas

    def fit_data(self, simulation, spacing_type=None, dual=False, taus=None, kappas_bar=None, alphas=None, verbose=False):
        """Fits the reduced order model params to a given simulation data by linear interpolation between kappas.

        :param current: transport current of the model
        :param flux: internal flux of the model
        :param spacing_type: kappa distribution type (lin, log, invlog, exponential), defaults to None
        :param verbose: display all information, defaults to False
        """
        if self.N is None:
            if kappas_bar is not None:
                self.N = len(kappas_bar)
            elif taus is not None:
                self.N = len(taus)
            elif alphas is not None:
                self.N = len(alphas)
            else:
                raise ValueError(f'Undefined number of cells. Cant fit to data.')

        # fit taus
        if taus is None:
            taus = [1.5e-5] + (self.N - 1) * [2e-4]
            # taus =  self.N*[0.0] if tau_sweep is None else self.fit_taus(tau_sweep=tau_sweep, verbose=verbose)

        # fit kappas
        if kappas_bar is None:
            kappas_bar = self.fit_kappas(simulation, spacing_type=spacing_type, verbose=verbose)

        # fit alphas
        if alphas is None:
            kappas_bar, alphas = self.fit_alphas(simulation, kappas_bar, verbose=verbose, dual=dual)

        # save fitted params
        self.simulation = simulation
        self.update(len(taus), taus, kappas_bar, alphas)

    def __closest_matches(self, dataset, values):
        return [min(range(len(dataset)), key=lambda i: abs(dataset[i] - value)) for value in values]

    def __flux_multicells(self, Irev):
        flux = 0
        for k in range(self.N):
            flux += self.alphas[k] * self.L0 * Irev[k]
        return flux

    def __g(self, I, Irev_prev, kappa):
        """ 1Dvector/scalar play model for g (=Irev+Ieddy) with coercivity kappas """
        diff = I - Irev_prev
        diff_unit = diff / abs(diff) if diff != 0 else 0
        if abs(diff) >= kappa or diff * (diff - kappa * diff_unit) > 0:
            return I - kappa * diff_unit
        else:
            return Irev_prev

    def compute(self, time, I, sim=None, v=0):
        """ Computing flux and losses on a given interval """
        if (self.N or self.alphas[0] or self.kappas_bar[0] or self.taus[0]) is None:
            raise ValueError(f'Undefined parameters. Cant compute solution.')

        # Background field scaling
        kappa_scaling, fill_factor = [1.0, 0.5] if sim is None else self.fit_scaling(sim)

        # INPUT PARAMETERS -----------------------------------------------
        nbSteps = len(I)
        rel_tol = 1e-5  # Iterations are needed because of the field-dependent parameters.
        iter_max = 100
        weightsum = np.cumsum(self.alphas)
        # RESOLUTION -----------------------------------------------------
        flux = np.zeros((nbSteps,))
        Irev = np.zeros((nbSteps, self.N))
        Iirr = np.zeros((nbSteps, self.N))
        Ieddy = np.zeros((nbSteps, self.N))
        g = np.zeros((nbSteps, self.N))
        iter_tot = 0
        # Computation of the associated flux density
        for i in range(1, nbSteps):
            # Vector hysteresis
            flux[i] = flux[i - 1]
            dt = time[i] - time[i - 1]
            conv_crit = 2
            iter = 0
            # scalar play model
            while conv_crit > 1 and iter < iter_max:
                for k in range(self.N):
                    # update g = Irev + Ieddy with "scalar" play model
                    diff = I[i] - Irev[i - 1, k]
                    diff_unit = diff / abs(diff) if diff != 0 else 0
                    # overcritical
                    if abs(I[i]) >= kappa_scaling * self.Ic_0:
                        Irev[i, k] = Irev[i - 1, k] + (1 - fill_factor) * (I[i] - I[i - 1]) / (2 * weightsum[k])
                        Ieddy[i, k] = Ieddy[i - 1, k] + self.taus[k] * (Irev[i - 1, k] - Irev[i, k]) / (dt)
                        Iirr[i, k] = I[i] - Irev[i, k] - Ieddy[i, k]
                    # undercritical
                    else:
                        if abs(diff) >= kappa_scaling * self.kappas_bar[k] or diff * (diff - kappa_scaling * self.kappas_bar[k] * diff_unit) > 0:
                            g[i, k] = I[i] - kappa_scaling * self.kappas_bar[k] * diff_unit
                        else:
                            g[i, k] = Irev[i - 1, k]
                            # g[i,k] = self.__g(I[i], Irev[i-1,k], self.kappas_bar[k])
                        # update irreversible
                        Iirr[i, k] = I[i] - g[i, k]
                        # update eddy
                        Ieddy[i, k] = self.taus[k] / (dt + self.taus[k]) * (g[i, k] - Irev[i - 1, k])
                        # calculate reversible
                        Irev[i, k] = g[i, k] - Ieddy[i, k]

                flux_new = self.__flux_multicells(Irev[i, :])
                conv_crit = abs((flux_new - flux[i]) / (flux[i] + 1e-10)) / rel_tol
                flux[i] = flux_new
                iter += 1
            iter_tot += iter
            if iter == iter_max:
                print(f"Step {i} reached maximum of {iter} iterations.")
        if v == 1:
            print(f'Finished with {iter_tot / nbSteps} iterations per step on average.')
        # POST-PROCESSING ------------------------------------------------
        # Computation of the stored and dissipated power quantities
        p_ohm_fil, p_ohm_mat = self.ohmic_loss(I, simulation=sim)
        p_rev = 0
        p_irr = 0
        p_eddy = 0
        for k in range(self.N):
            dfluxdt = self.L0 * np.concatenate(([0], np.diff(Irev[:, k]) / np.diff(time)))  # np.gradient(Irev[:,k], time)
            # Instantaneous power. Stored: \sum_k ( alpha_k * Irev_k * dphi_k/dt ). Dissipated: \sum_k ( alpha_k * Iirr_k * dphi_k/dt )
            p_rev += self.alphas[k] * Irev[:, k] * dfluxdt  # See Appendix B.2. of J. Dular's thesis (https://orbi.uliege.be/handle/2268/298054)
            p_irr += self.alphas[k] * Iirr[:, k] * dfluxdt
            p_eddy += self.alphas[k] * Ieddy[:, k] * dfluxdt

        # Per cycle loss
        p_fil_cycle = None
        p_mat_cycle = None
        if sim and sim.f:
            t_end = np.array(time)[-1]
            # Cumulative integration over one cycle
            x_init = self.__closest_matches(time, [t_end - 1 / (sim.f)])[0]
            cumul_p_fil = integrate.cumulative_trapezoid(p_irr + p_ohm_fil, time, initial=0)
            cumul_p_mat = integrate.cumulative_trapezoid(p_eddy + p_ohm_mat, time, initial=0)
            p_fil_cycle = (cumul_p_fil[-1] - cumul_p_fil[x_init])
            p_mat_cycle = (cumul_p_mat[-1] - cumul_p_mat[x_init])

        return flux, p_irr + p_ohm_fil, p_eddy + p_ohm_mat, p_fil_cycle, p_mat_cycle


class PostProcess:
    """
    This class loads and stores the data from a simulation and can apply various postprocessing operations on the data.
    The simulation data is saved as a SimulationData object.
    """

    def __init__(self, fdm, model_data_output_path) -> None:
        self.fdm = fdm
        self.model_data_output_path = model_data_output_path
        self.geometry_name = f"Geometry_{self.fdm.run.geometry}"
        self.mesh_name = f"Mesh_{self.fdm.run.mesh}"
        self.solution_name = f"Solution_{self.fdm.run.solution}"

        self.simulation_data = SimulationData(model_data_output_path, self.geometry_name, self.mesh_name, self.solution_name, self.fdm)
        self.flux_model = ROHFmodel(N=self.fdm.magnet.postproc.plot_flux.rohf_N,
                                    spacing_type=self.fdm.magnet.postproc.plot_flux.rohf_kappa_spacing,
                                    simulation=self.simulation_data,
                                    filename=self.fdm.magnet.postproc.plot_flux.rohf_file,
                                    filepath=os.path.join(self.model_data_output_path, self.geometry_name),
                                    ) if self.fdm.magnet.postproc.plot_flux.rohf else None
        self.plotter = PlotPythonConductorAC(fdm, model_data_output_path, self.simulation_data, self.flux_model)
        self.report = PDFreport(filepath=model_data_output_path,
                                title=f'{self.fdm.run.geometry}_{self.fdm.run.mesh}_{self.fdm.run.solution}' if self.fdm.magnet.postproc.generate_report else None)

    def rohf_background(self):
        """ Test for ROHFmodel background field scaling """
        if steammaterials_flag or self.flux_model is None:
            # skip
            return
        matpath = os.path.join(os.path.dirname(steammaterials.__file__), 'CFUN')

        Ic_0 = 2900
        B_background = self.simulation_data.B_background

        # analytic self-field on strand boundary
        CATI_surf = 54 * np.pi * (4.5e-05) ** 2  # 54 filament surface in CATI paper strand
        Jc_0 = Ic_0 / CATI_surf
        B_space = np.linspace(0, 14, 5000)
        # Determine ambient field with no background field and max current density
        numpy2d_Bspace = np.vstack((1.9 * np.ones(B_space.shape), B_space, 1 * np.ones(B_space.shape)))  # T, B, A
        nbti_space = steammaterials.SteamMaterials("CFUN_IcNbTi_v1", numpy2d_Bspace.shape[0], numpy2d_Bspace.shape[1], matpath)
        jc_nbti_space = nbti_space.evaluate(numpy2d_Bspace)
        B_ambient = B_space[np.abs(jc_nbti_space - Jc_0).argmin()]
        print("B_ambient: " + str(B_ambient) + " [T]")

        # Determine absolute field value which the strand experiences
        B_abs = B_background + (B_ambient / B_background) * B_ambient if B_background > B_ambient else B_background + B_ambient
        # material Jc(B) fit
        jc_nbti = steammaterials.SteamMaterials("CFUN_IcNbTi_v1", 3, 1, matpath)
        jc_Babs = jc_nbti.evaluate((1.9, B_abs, 1))[0]

        # PLOT
        plt.figure()
        plt.plot(B_space, jc_nbti_space, label='CFUN_IcNbTi_v1')
        plt.scatter((B_ambient, B_abs), (Jc_0, jc_Babs), label='Jc(B)', )
        plt.title('Nb-Ti')
        plt.grid()
        plt.legend()
        plt.xlabel("Field (T)")
        plt.ylabel("jc (A/m2)")

        kappa_scaling = jc_Babs / Jc_0
        print(" Kappa scaling factor: " + str(kappa_scaling))
        self.flux_model.kappa_scaling = kappa_scaling
        plt.show()

    def internal_flux(self):
        """ Postproc routine for the internal flux and the related ROHF model """
        self.plotter.plot_transport()
        self.plotter.plot_flux()
        self.plotter.plot_power_loss()
        self.plotter.plot_resistivity()

        # self.flux_model.export_params(filepath=os.path.join(self.model_data_output_path, self.geometry_name), filename="TESTFILE")
        # self.flux_model.import_params(filepath=os.path.join(self.model_data_output_path, self.geometry_name), filename="TESTFILE")

        if self.report and self.flux_model:
            self.report.add_table('ROHF parameters', self.flux_model.params())

    def instantaneous_loss(self):
        print("Total loss: ", self.simulation_data.cumulative_power["TotalLoss"].iloc[-1])
        print("Total filament loss: ", self.simulation_data.cumulative_power["FilamentLoss"].iloc[-1])
        print("Total coupling loss: ", self.simulation_data.cumulative_power["CouplingLoss"].iloc[-1])
        print("Total eddy loss: ", self.simulation_data.cumulative_power["EddyLoss"].iloc[-1])
        plot_options = self.fdm.magnet.postproc.plot_instantaneous_power
        self.plotter.plot_instantaneous_power(
            show=plot_options.show,
            title=plot_options.title,
            save_plot=plot_options.save,
            save_folder_path=os.path.join(self.model_data_output_path, self.geometry_name, self.mesh_name, self.solution_name),
            save_file_name=plot_options.save_file_name,
            overwrite=self.fdm.run.overwrite
        )

    def plot_impedance(self):
        """ Plots the impedance estimate and transport quantities """
        conductor = self.fdm.conductors[self.fdm.magnet.solve.conductor_name]
        frequency_range = np.logspace(-2, 6, 1000)

        self.simulation_data.plot_transport()

        #### VOLTAGE ####
        # self.simulation_data.plot_phaseshift_voltage(90)
        # self.simulation_data.plot_voltage_decomp()

        #### IMPEDANCE ####
        self.simulation_data.compare_impedance(frequency_range)


class BatchPostProcess:
    """
    This class loads and stores data from multiple simulations (+flux models) and performs operations on them
    """

    def __init__(self, fdm, lossMap_gridData_folder, inputs_folder_path, outputs_folder_path) -> None:
        self.fdm = fdm
        self.inputs_folder_path = inputs_folder_path  # This is the path to the folder where the input data is stored
        self.model_data_output_path = outputs_folder_path  # This is the path to the folder where the model output data is stored (e.g. geometries)
        self.outputs_folder_path = os.path.join(outputs_folder_path, fdm.magnet.postproc.output_folder)  # This is the path to the folder where the postprocessed data is written
        self.simulation_collection = self.retrieve_simulation_data()
        self.fluxmodel_collection = self.retrieve_fluxmodel_data()
        self.plotter = BatchPlotPythonConductorAC(
            fdm=self.fdm, lossMap_gridData_folder=lossMap_gridData_folder, outputs_folder_path=outputs_folder_path,
            simulation_collection=self.simulation_collection, fluxmodel_collection=self.fluxmodel_collection
        )
        self.report = PDFreport(filepath=outputs_folder_path, title='BatchPostprocReport') if self.fdm.magnet.postproc.generate_report else None
        self.avg_simulation_time = np.mean([sd.simulation_time for sd in self.simulation_collection])
        self.total_simulation_time = np.sum([sd.simulation_time for sd in self.simulation_collection])
        print('Number of simulations considered: ', len(self.simulation_collection))
        print('Average simulation time: ', self.avg_simulation_time, 's')
        print('Total simulation time: ', self.total_simulation_time, 's')

    def get_base_sim(self):
        """ Find simulation with lowest frequency and highest amplitude"""
        base_sim = self.simulation_collection[0]
        for sim in self.simulation_collection[1:]:
            if sim.f <= base_sim.f and sim.I_amp >= base_sim.I_amp: base_sim = sim
        return base_sim

    def rohf_on_grid(self):
        """ Checks the fit of an existing model to a grid of simulations or fits a new ROHF model to a grid of simulations """

        optimized_rohf = None
        if self.fdm.magnet.postproc.batch_postproc.rohf_on_grid.fit_rohf:

            tausweep_IIC = self.fdm.magnet.postproc.batch_postproc.rohf_on_grid.fit_rohf_tausweep_IIC
            N = self.fdm.magnet.postproc.batch_postproc.rohf_on_grid.fit_rohf_N
            print("FITTING ROHF TO SIMULATIONS:")
            print("-> number of cells N = " + str(N))
            print("-> tausweep IIC = " + str(tausweep_IIC))

            taus = self.optimize_taus(tausweep_IIC=tausweep_IIC, N=N)
            kappas, alphas = self.optimize_kappas(base_taus=taus, N=N)
            # export resulting ROHF
            optimized_rohf = ROHFmodel(N=N, kappas_bar=kappas, alphas=alphas, taus=taus)
            print(optimized_rohf.taus)
            print(optimized_rohf.alphas)
            print(optimized_rohf.kappas_bar)
            self.plotter.plot_flux_time_triplet(flux_model=optimized_rohf, title='Flux in time triplet')

            optimized_rohf.export_params(self.outputs_folder_path, 'optimized')

            if self.report:
                self.report.filename = 'ROHFoptimization.pdf'
                self.report.add_table('Parameter results', optimized_rohf.params())

        if self.fdm.magnet.postproc.batch_postproc.rohf_on_grid.produce_error_map:
            flux_model = optimized_rohf if optimized_rohf is not None else self.fluxmodel_collection[0]
            self.plotter.create_errorMap(flux_model=flux_model)

            if self.report:
                self.report.add_table('Error map parameter', optimized_rohf.params())

    def optimize_kappas(self, N, base_kappas=None, base_taus=None):

        base_sim = self.get_base_sim()
        if base_kappas is None:
            base_kappas = np.linspace(0, base_sim.I_amp, N + 1)[:-1]  # linear spaced over max I range
            # base_kappas = np.geomspace(1, base_sim.I_amp, N+1)[:-1]-1
        if base_taus is None:
            base_taus = np.zeros(N)
        baseROHF = ROHFmodel(simulation=base_sim, taus=base_taus, kappas_bar=base_kappas)

        print(" RUNNING KAPPA & WEIGHT OPTIMIZATION ...")

        # ERROR -------------------------------------------------------------------------------------------------------------------------------
        def fit_error_kappas(normalized_kappas):
            # static alpha fit based on recursion of kappas -> piecewise linear fit
            baseROHF.fit_data(taus=base_taus, simulation=base_sim, kappas_bar=normalized_kappas * base_sim.I_amp)
            flux_error = 0
            for simulation in self.simulation_collection:
                time, I, flux = simulation.halfcycle(current=True)
                rohf_flux, _, _, _, _ = baseROHF.compute(time, I)
                flux_error = flux_error + sum(np.diff(time, prepend=0) * np.absolute(flux - rohf_flux))
            return flux_error
            # BOUNDARIES --------------------------------------------------------------------------------------------------------------------------

        bounds = optimize.Bounds(lb=0, ub=1, keep_feasible=True)
        # OPTIMIZATION 1: KAPPA DISTRIBUTION & WEIGHTS ----------------------------------------------------------------------------------------
        kappa_optimization = optimize.minimize(fit_error_kappas, method='L-BFGS-B', x0=base_kappas / base_sim.I_amp, bounds=bounds, options={"disp": True})
        optimized_kappas = np.sort(kappa_optimization.x * base_sim.I_amp)
        print(" --> OPTIMIZED KAPPAS: " + str(optimized_kappas))

        # ERROR -------------------------------------------------------------------------------------------------------------------------------
        def fit_error_alphas(alphas):
            baseROHF.alphas = alphas
            flux_error = 0
            for simulation in self.simulation_collection:
                time, I, flux = simulation.halfcycle(current=True)
                rohf_flux, _, _, _, _ = baseROHF.compute(time, I)
                flux_error = flux_error + sum(abs(np.diff(I, prepend=0)) * np.diff(time, prepend=0) * np.absolute(flux - rohf_flux))
            return flux_error
            # BOUNDARIES --------------------------------------------------------------------------------------------------------------------------

        bounds = optimize.Bounds(lb=0, ub=10, keep_feasible=True)
        # OPTIMIZATION 2: WEIGHTS ONLY ---------------------------------------------------------------------------------------------------------
        alpha_optimization = optimize.minimize(fit_error_alphas, method='L-BFGS-B', x0=baseROHF.alphas, bounds=bounds, options={"disp": True})
        optimized_alphas = alpha_optimization.x
        print(" --> OPTIMIZED ALPHAS: " + str(optimized_alphas))

        return [optimized_kappas, optimized_alphas]

    def optimize_taus(self, tausweep_IIC, N, base_taus=None, base_kappas=None):

        force_tau1_zero = False
        base_sim = self.get_base_sim()

        amplitudes = sorted(list(set([sim.I_amp for sim in self.simulation_collection])))
        tausweep_I = min(amplitudes, key=lambda x: abs(x - tausweep_IIC * amplitudes[-1]))  # get grid current closest to tausweep IIC
        tau_sweep = [sim for sim in self.simulation_collection if sim.I_amp == tausweep_I]  # get asociated sweep

        freq = sorted(list(set([sim.f for sim in tau_sweep])))
        if base_kappas is None:
            base_kappas = np.linspace(0, amplitudes[-1], N + 1)[:-1]  # start with linear spaced over max_I range
        if base_taus is None:
            # Estimate initial tau values with lowpass approximation
            loss = [sim.total_power_per_cycle['FilamentLoss'] for sim in tau_sweep]
            resamp_freq = np.linspace(freq[0], freq[-1], 500)
            resamp_loss = np.interp(resamp_freq, freq, loss)
            idx = np.abs(loss - resamp_loss[0] * 0.707).argmin()  # -3dB mark -> crit freq
            tau_star = 1 / (2 * np.pi * resamp_freq[idx])  # f_c of lowpass
            base_taus = base_kappas / (N * tausweep_I) * tau_star  # scale estimate down w. cell level

        baseROHF = ROHFmodel(N=N, simulation=base_sim, kappas_bar=base_kappas, taus=base_taus)
        self.plotter.plot_percycle_losses(sims=tau_sweep, flux_model=baseROHF, title='STARTING TAUS')

        if True:
            print(" RUNNING TAU OPTIMIZATION ... ")

            # ERROR -------------------------------------------------------------------------------------------------------------------------------
            def fit_error(taus):
                baseROHF.taus = np.concatenate(([0], taus)) if force_tau1_zero else taus
                error = 0
                for idx, sim in enumerate(self.simulation_collection):
                    time, I, flux = sim.halfcycle(current=True)
                    flux_rohf, fill_rohf, eddy_rohf, _, _ = baseROHF.compute(time=time, I=I)
                    # errors[idx] = abs(sim.total_power_per_cycle['EddyLoss']-eddy_pc_rohf) + abs(sim.total_power_per_cycle['FilamentLoss']-fill_pc_rohf)
                    error = error + sum(abs((sim.power['EddyLoss'][:len(I)] - eddy_rohf)) / np.mean(sim.power['EddyLoss']) + abs((sim.power['FilamentLoss'][:len(I)] - fill_rohf)) / np.mean(sim.power['FilamentLoss']))
                return error
                # BOUNDARIES --------------------------------------------------------------------------------------------------------------------------

            bounds = optimize.Bounds(lb=0.0, ub=0.1)
            # OPTIMIZATION ------------------------------------------------------------------------------------------------------------------------
            optimization = optimize.minimize(fit_error, x0=base_taus, bounds=bounds, options={'disp': True})  # , method='SLSQP'
            optimized_taus = np.concatenate(([0], optimization.x[1:])) if force_tau1_zero else optimization.x

            print(" --> OPTIMIZED TAUS: " + str(optimized_taus))
            baseROHF.fit_data(simulation=base_sim, spacing_type='linear', taus=np.sort(optimized_taus))
            self.plotter.plot_percycle_losses(sims=tau_sweep, flux_model=baseROHF, title='OPTIMIZED TAUS')
            # plt.show()
            return np.sort(optimized_taus)

        else:
            return baseROHF.taus

    def retrieve_simulation_data(self):
        """
        This function iterates over the input CSV-file (specifying which simulations to postprocess) and returns a list of SimulationData objects
        containing all the simulation data. If no CSV-file is specified, the data from the single simulation specified in the input YAML-file is returned.
        """

        simulations_csv = self.fdm.magnet.postproc.batch_postproc.simulations_csv
        if simulations_csv is not None:
            try:
                self.simulations_csv = pd.read_csv(os.path.join(self.inputs_folder_path, f'{simulations_csv}.csv'))  # Read the csv file with the input data
            except:
                raise FileNotFoundError(f'No csv file with the name {simulations_csv}.csv was found in the inputs folder.')

        if self.simulations_csv is not None:
            simulationCollection = []
            for index, row in self.simulations_csv.iterrows():
                if pd.isna(row['input.run.geometry']) and pd.isna(row['input.run.mesh']) and pd.isna(row['input.run.solution']):
                    continue
                geometry_name = 'Geometry_' + str(row['input.run.geometry'])
                mesh_name = 'Mesh_' + str(row['input.run.mesh'])

                if isinstance(row['input.run.solution'], float) and row['input.run.solution'].is_integer():
                    solution_name = 'Solution_' + str(int(row['input.run.solution']))
                else:
                    solution_name = 'Solution_' + str(row['input.run.solution'])

                # Check if the row refers to a valid simulation by checking if the solution folder exists:
                # solution_folder = os.path.join(os.getcwd(), 'tests', '_outputs', self.fdm.general.magnet_name, geometry_name, mesh_name, solution_name)
                solution_folder = os.path.join(self.model_data_output_path, geometry_name, mesh_name, solution_name)
                if os.path.exists(solution_folder):  # If the solution folder exists, add the simulation to the simulationCollection
                    sd = SimulationData(self.model_data_output_path, geometry_name, mesh_name, solution_name, self.fdm)
                    if sd.simulation_time is not None:  # Only add the simulation if it has finished running (and therefore has written the simulation time to a file)
                        simulationCollection.append(sd)
        else:
            simulationCollection = [SimulationData(self.model_data_output_path, 'Geometry_' + self.fdm.run.geometry, 'Mesh_' + self.fdm.run.mesh, 'Solution_' + self.fdm.run.solution)]

        return self.sort_simulationCollection(self.filter_simulationCollection(simulationCollection))

    def filter_simulationCollection(self, simulationCollection):
        """
        This function is used to filter the simulationCollection based on the filter criterion specified in the yaml input file.
        An example of a filter criterion is '<<solve.source_parameters.sine.frequency>> == 18', which will disregard all simulations with frequency != 18Hz.
        """
        if self.fdm.magnet.postproc.batch_postproc.filter.apply_filter:
            filter_criterion = self.fdm.magnet.postproc.batch_postproc.filter.filter_criterion
            class_params = re.findall('<<(.*?)>>', filter_criterion)
            for cp in class_params:
                filter_criterion = filter_criterion.replace(f"<<{cp}>>", 'sd.' + cp)
            filtering_function = eval(f'lambda sd: {filter_criterion}')
            return list(filter(filtering_function, simulationCollection))
        else:
            return simulationCollection

    def sort_simulationCollection(self, simulationCollection):
        """
        This function is used to sort the simulationCollection based on the sort key specified in the yaml input file.
        An example of a sort key is 'solve.source_parameters.sine.frequency', which will sort the simulations based on frequency.
        """
        if self.fdm.magnet.postproc.batch_postproc.sort.apply_sort:
            sorting_function = eval(f'lambda sd: sd.{self.fdm.magnet.postproc.batch_postproc.sort.sort_key}')
            return sorted(simulationCollection, key=sorting_function)
        else:
            return simulationCollection

    def retrieve_fluxmodel_data(self):
        """ Analog function to retrieve_simulation_data but for fluxmodels (Glock-Thesis). """
        fluxmodels_csv = self.fdm.magnet.postproc.batch_postproc.fluxmodels_csv
        if fluxmodels_csv is not None:
            try:
                self.fluxmodels_csv = pd.read_csv(os.path.join(self.inputs_folder_path, f'{fluxmodels_csv}.csv'))  # Read the csv file with the input data
            except:
                raise FileNotFoundError(f'No csv file with the name {fluxmodels_csv}.csv was found in the inputs folder.')

        fluxmodelCollection = []
        if fluxmodels_csv is not None:
            for index, row in self.fluxmodels_csv.iterrows():
                if pd.isna(row['input.run.geometry']) and pd.isna(row['input.run.mesh']) and pd.isna(row['input.run.solution']):
                    continue
                geometry_name = 'Geometry_' + str(row['input.run.geometry'])
                parameterfile_path = os.path.join(self.model_data_output_path, geometry_name)
                parameterfile_name = str(row['input.run.parameterfile'])
                label_name = str(row['input.run.label'])

                fluxmodelCollection.append(ROHFmodel(filepath=parameterfile_path, filename=parameterfile_name, label=label_name))

        return fluxmodelCollection


import os
import pickle
import logging

import gmsh
import numpy as np

from fiqus.data import RegionsModelFiQuS
from fiqus.utils.Utils import GmshUtils, FilesAndFolders
from fiqus.data.RegionsModelFiQuS import RegionsModel

from fiqus.geom_generators.GeometryConductorAC_Strand import TwistedStrand, Line

occ = gmsh.model.occ

logger = logging.getLogger('FiQuS')
    
class Mesh:
    def __init__(self, fdm, verbose=True):
        """
        A base-class used to manage the mesh for the CAC Strand and Rutherford models.

        :ivar fdm: The fiqus data model for input parameters.
        :vartype fdm: dict
        :ivar cacdm: The magnet section from the fdm.
        :vartype cacdm: object
        :ivar mesh_folder: The path to the folder where the mesh files are stored.
        :vartype mesh_folder: str
        :ivar magnet_name: The name of the magnet model.
        :vartype magnet_name: str
        :ivar mesh_file: The path to the .msh file for the mesh.
        :vartype mesh_file: str
        :ivar regions_file: The path to the .regions file for the mesh.
        :vartype regions_file: str
        :ivar verbose: If True, the class will print additional information during execution.
        :vartype verbose: bool
        :ivar gu: An instance of the GmshUtils class for managing the gmsh utility.
        :vartype gu: GmshUtils
        """
        self.fdm = fdm
        self.cacdm = fdm.magnet
        self.mesh_folder = os.path.join(os.getcwd())
        self.magnet_name = fdm.general.magnet_name
        self.mesh_file = os.path.join(self.mesh_folder, f"{self.magnet_name}.msh")
        self.regions_file = os.path.join(self.mesh_folder, f"{self.magnet_name}.regions")
        self.verbose = verbose
        self.gu = GmshUtils(self.mesh_folder, self.verbose)
        self.gu.initialize(verbosity_Gmsh=fdm.run.verbosity_Gmsh)

        gmsh.option.setNumber("General.Terminal", verbose)
    
    def save_mesh(self, gui: bool = False):
        """
            Saves the mesh to a .msh file. If gui is True, the mesh is also loaded in the gmsh GUI.
        """
        gmsh.write(self.mesh_file)
        if gui:
            self.gu.launch_interactive_GUI()
        else:
            if gmsh.isInitialized():
                gmsh.clear()
                gmsh.finalize()

    def load_mesh(self, gui : bool = False):
        """
            Loads a previously generated mesh.
        """
        gmsh.clear()
        gmsh.open(self.mesh_file)

        if gui:
            self.gu.launch_interactive_GUI()

class StrandMesh(Mesh):
    """
    A subclass of Mesh that handles mesh generation for the twisted strand cross-section geometries.

    :ivar geometry_class: An instance of the TwistedStrand class that stores all the information about the model structure. Initially set to None and should be loaded or assigned before mesh generation methods are called.
    :vartype geometry_class: TwistedStrand

    :param fdm: The fiqus data model, providing all the necessary data and parameters for mesh generation.
    :type fdm: object
    :param verbose: A boolean flag that indicates whether additional information should be printed to the console during operation. Defaults to True.
    :type verbose: bool, optional
    """
    def __init__(self, fdm, verbose=True):
        super().__init__(fdm, verbose)
        self.geometry_class : TwistedStrand = None # Class from geometry generation step, used to store all information about the model structure.

    def filament_field(self):
        """
        Generates the filament mesh size field.

        The standard mesh size field is linearly interpolated between the mesh size at the filament boundary and the center. 
        Amplitude dependent meshing is applied if the field is expected to only penetrate a few elements in the filaments, adjusting the mesh size near the filament edges to capture the field penetration region accurately.

        :return: The filament mesh size field tag.
        :rtype: int
        """
        # Define some constants
        # Choose the largest distance from the center of mass of the filaments to the filament boundary as the filament radius.
        # This approximation is done to account for the fact that the filaments may not be circular, but can have any shape.
        filament_rad = 0 
        for layer in self.geometry_class.filaments:
            for filament in layer:
                center_of_mass = gmsh.model.occ.get_center_of_mass(2, filament.surface_tag)
                for curve in filament.boundary_curves:
                    r = max([np.linalg.norm(np.array(center_of_mass) - curve.P1.pos), np.linalg.norm(np.array(center_of_mass) - curve.P2.pos)])
                    if r > filament_rad:
                        filament_rad = r
        meshSize_at_filament_boundary = filament_rad * self.cacdm.mesh.filaments.boundary_mesh_size_ratio
        meshSize_at_filament_center = filament_rad * self.cacdm.mesh.filaments.center_mesh_size_ratio
        filament_edges = sum([filament.boundary_curves for layer in self.geometry_class.filaments for filament in layer], [])


        amplitude_dependent_meshing, field_penetration_distance = self.evaluate_amplitude_dependent_meshing().values()
        desired_elements_in_field_penetration_region = self.cacdm.mesh.filaments.desired_elements_in_field_penetration_region

        frequency_dependent_meshing, skin_depth = self.evaluate_frequency_dependent_meshing().values()
        if frequency_dependent_meshing:
            filaments_in_skindepth = [filament for layer in self.geometry_class.filaments for filament in layer if np.linalg.norm(filament.center_point.pos) > self.geometry_class.matrix[-1].rad - 2*skin_depth]
        else:
            filaments_in_skindepth = [filament for layer in self.geometry_class.filaments for filament in layer]

        if amplitude_dependent_meshing:
            # The field only penetrates a few elements. We should thus decrease the mesh size close to the filament edges. 
            # min_mesh_size will be overwritten, while meshSize_at_filament_center will remain. 

            circular_filament_fields = []
            nodesPerFilament = 2*np.pi*filament_rad / ((field_penetration_distance)/desired_elements_in_field_penetration_region) # We place nodes for meshing at the filament boundary for more efficient meshing (transfinite meshing)

            for filament in filaments_in_skindepth:
                circular_field = gmsh.model.mesh.field.add("Ball")
                gmsh.model.mesh.field.setNumber(circular_field, "Radius", (filament_rad-field_penetration_distance)*0.85 )
                gmsh.model.mesh.field.setNumber(circular_field, "XCenter", filament.center_point.pos[0])
                gmsh.model.mesh.field.setNumber(circular_field, "YCenter", filament.center_point.pos[1])
                gmsh.model.mesh.field.setNumber(circular_field, "Thickness", (filament_rad-field_penetration_distance)*0.15)
                gmsh.model.mesh.field.setNumber(circular_field, "VIn", meshSize_at_filament_center)
                gmsh.model.mesh.field.setNumber(circular_field, "VOut", (field_penetration_distance)/desired_elements_in_field_penetration_region )
                circular_filament_fields.append(circular_field)

                for circle_arc_tag in [CA.tag for CA in filament.boundary_curves]:
                    gmsh.model.mesh.setTransfiniteCurve(circle_arc_tag, int(nodesPerFilament/len(filament.boundary_curves)), "Progression", 1)

            filament_fields = gmsh.model.mesh.field.add("Max")
            gmsh.model.mesh.field.setNumbers(filament_fields, "FieldsList", circular_filament_fields)

            self.cacdm.mesh.filaments.boundary_mesh_size_ratio = ((field_penetration_distance)/desired_elements_in_field_penetration_region)/filament_rad

        
        else:
            # Generalized to arbitrary filament shapes (not only circular)
            f_dist_boundary = gmsh.model.mesh.field.add("Distance")
            gmsh.model.mesh.field.setNumbers(f_dist_boundary, "CurvesList", [edge.tag for edge in filament_edges])
            
            mass_centers = []
            for layer in self.geometry_class.filaments:
                for filament in layer:
                    x,y,z = gmsh.model.occ.get_center_of_mass(2, filament.surface_tag)
                    mass_center_tag = gmsh.model.occ.addPoint(x, y, z)
                    mass_centers.append(mass_center_tag)
            gmsh.model.occ.synchronize()

            f_dist_center = gmsh.model.mesh.field.add("Distance")
            gmsh.model.mesh.field.setNumbers(f_dist_center, "PointsList", mass_centers)

            # Linearly interpolate between the mesh size at the filament center and the mesh size at the filament boundary.
            filament_fields = gmsh.model.mesh.field.add("MathEval") 
            gmsh.model.mesh.field.setString(
                filament_fields, 
                "F", 
                f"( {meshSize_at_filament_boundary} * F{f_dist_center} + {meshSize_at_filament_center} * F{f_dist_boundary} ) / ( F{f_dist_center} + F{f_dist_boundary} )"
            )


        # Restrict the filament fields to the filaments
        filament_mesh_field = gmsh.model.mesh.field.add("Restrict")
        gmsh.model.mesh.field.setNumber(filament_mesh_field, "InField", filament_fields)

        
        if frequency_dependent_meshing:
            gmsh.model.mesh.field.setNumbers(filament_mesh_field, "SurfacesList", [filament.surface_tag for filament in filaments_in_skindepth])
        else:
            gmsh.model.mesh.field.setNumbers(filament_mesh_field, "SurfacesList", [filament.surface_tag for layer in self.geometry_class.filaments for filament in layer] + [hole.surface_tag for layer in self.geometry_class.filament_holes for hole in layer])

        return filament_mesh_field

    def matrix_threshold_to_filament_field(self, filaments):
        """
        Adjusts the matrix mesh size based on the distance from the filaments to improve mesh quality.

        This method creates a mesh size field that gradually refines the mesh in the matrix as it approaches the filaments. 
        Filaments typically have a finer mesh than the matrix, and a large gradient in mesh size at their boundary can degrade the mesh quality. 
        The method interpolates the mesh size from a minimum value at the filament boundary to a maximum value at a specified distance from the filaments, based on configuration parameters in the FDM.

        :param filaments: A list of filament surface objects for which the boundary mesh size adjustments are to be made.
        :type filaments: list[list[object]]

        :return: The tag of the created mesh size field, which is restricted to the matrix domain.
        :rtype: int
        """
        matrix = self.geometry_class.matrix
        matrix_mesh_options = self.cacdm.mesh.matrix

        filament_edges = sum([[CA.tag for CA in filament.boundary_curves] for filament in filaments], []) # List of tags of all filament boundaries
        # Distance to filament boundaries field:
        distance_to_filament_field = gmsh.model.mesh.field.add("Distance")
        gmsh.model.mesh.field.setNumbers(distance_to_filament_field, "CurvesList", filament_edges)

        # Choose the largest distance from the center of mass of the filaments to the filament boundary as the filament radius.
        # This approximation is done to account for the fact that the filaments may not be circular, but can have any shape.
        filament_rad = 0 
        for layer in self.geometry_class.filaments:
            for filament in layer:
                center_of_mass = gmsh.model.occ.get_center_of_mass(2, filament.surface_tag)
                for curve in filament.boundary_curves:
                    r = max([np.linalg.norm(np.array(center_of_mass) - curve.P1.pos), np.linalg.norm(np.array(center_of_mass) - curve.P2.pos)])
                    if r > filament_rad:
                        filament_rad = r

        filament_boundary_meshSize = filament_rad * self.cacdm.mesh.filaments.boundary_mesh_size_ratio
        
        # Linarily interpolate the mesh size from the filament boundaries to a specified distance from filaments.
        threshold_to_filament_field = gmsh.model.mesh.field.add("Threshold")
        gmsh.model.mesh.field.setNumber(threshold_to_filament_field, "InField", distance_to_filament_field)
        gmsh.model.mesh.field.setNumber(threshold_to_filament_field, "SizeMin", filament_boundary_meshSize)
        gmsh.model.mesh.field.setNumber(threshold_to_filament_field, "SizeMax", filament_rad*matrix_mesh_options.mesh_size_matrix_ratio_inner)
        gmsh.model.mesh.field.setNumber(threshold_to_filament_field, "DistMin", 0)
        gmsh.model.mesh.field.setNumber(threshold_to_filament_field, "DistMax", filament_rad*matrix_mesh_options.interpolation_distance_from_filaments_ratio)
        gmsh.model.mesh.field.setNumber(threshold_to_filament_field, "StopAtDistMax", 1)
        # gmsh.model.mesh.field.setNumber(threshold_to_filament_field, "Sigmoid", 0)

        meshSize_by_filament_field = gmsh.model.mesh.field.add("Restrict") # Restrict the mesh to the matrix
        gmsh.model.mesh.field.setNumber(meshSize_by_filament_field, "InField", threshold_to_filament_field)
        gmsh.model.mesh.field.setNumbers(meshSize_by_filament_field, "SurfacesList", [partition.surface_tag for partition in matrix])
        # matrix_mesh_size_fields.append(meshSize_by_filament_field)
        return meshSize_by_filament_field

    def evaluate_amplitude_dependent_meshing(self):
        """
        Evaluates the necessity for amplitude-dependent meshing based on the applied magnetic field's amplitude.

        This method assesses whether the amplitude of the applied magnetic field is low enough that it only partially penetrates the filaments. 
        In such cases, a finer mesh is required in the field penetration region to accurately capture field variations, while a coarser mesh can be used in the filament center where the field does not penetrate. 
        It calculates the field penetration distance and determines whether an adjustment in the mesh is necessary to ensure accurate simulation results.

        :return: A dictionary containing:
            - "amplitude_dependent_meshing" (bool): Indicates whether amplitude-dependent meshing is needed based on the magnetic field's penetration.
            - "field_penetration_distance" (float): The estimated distance the magnetic field penetrates into the filaments, used to adjust the mesh size accordingly.
        :rtype: Dict[str, Union[bool, float]]
        """
        if self.cacdm.mesh.filaments.amplitude_dependent_scaling and self.cacdm.solve.source_parameters.sine.current_amplitude == 0 and not self.cacdm.geometry.io_settings.load.load_from_yaml:
            filament_radius = self.geometry_class.filaments[0][0].rad
            min_meshSize = filament_radius * self.cacdm.mesh.filaments.boundary_mesh_size_ratio
            meshSize_at_filament_center = filament_radius * self.cacdm.mesh.filaments.center_mesh_size_ratio

            applied_field_amplitude = self.cacdm.solve.source_parameters.sine.field_amplitude
            applied_field_amplitude = max(applied_field_amplitude, 0.01) # Limit the amplitude to 0.01 T to avoid excessively fine mesh (Could be removed?).
            jc = 3e9 #self.cacdm.solve.material_properties.superconducting.jc.constant
            mu0 = np.pi*4e-7

            field_penetration_distance = applied_field_amplitude/(mu0*jc) * self.cacdm.mesh.filaments.field_penetration_depth_scaling_factor # Estimate the field penetration distance

            # The mesh size in the filaments can be written as mesh_size(x) = a*x+b, with x being the distance from the filament boundary and a,b defined as:
            a = (meshSize_at_filament_center - min_meshSize) / filament_radius
            b = min_meshSize
            # The number of elements in the range x:[x0, x1] is approximated by the function:
            number_of_elements_in_range = lambda a, b, x0, x1 : np.log(a*x1+b)/a - np.log(a*x0+b)/a if a != 0 else (x1-x0)/b

            number_of_elements_penetrated_by_field = number_of_elements_in_range(a, b, 0, field_penetration_distance)
            desired_elements_in_field_penetration_region = self.cacdm.mesh.filaments.desired_elements_in_field_penetration_region

            if (number_of_elements_penetrated_by_field < desired_elements_in_field_penetration_region) and field_penetration_distance < filament_radius:
                logger.info(f"The magnetic field is expected to penetrate over approximately {round(number_of_elements_penetrated_by_field, 3)} elements in the filaments. To ensure that the field penetrates over at least {desired_elements_in_field_penetration_region} elements, the mesh in the filaments has been adjusted.")
                return dict(amplitude_dependent_meshing = True, field_penetration_distance = field_penetration_distance)
        
        return dict(amplitude_dependent_meshing = False, field_penetration_distance = 1e10)

    def evaluate_frequency_dependent_meshing(self):
        """
        Evaluates the need for frequency-dependent meshing based on the skin depth due to the applied magnetic field's frequency.

        This method determines if the frequency of the applied magnetic field is high enough to confine the currents to a thin layer near the surface of the strand. 
        A fine mesh is required in this region to capture current variations accurately, while a coarser mesh can suffice in the strand center. 
        The method calculates the skin depth using the field frequency and the matrix resistivity, then approximates the number of elements within this depth. 
        If the calculated number is less than desired, it indicates that frequency-dependent meshing adjustments are needed.

        :return: A dictionary containing:
            - "frequency_dependent_meshing" (bool): Indicates whether frequency-dependent meshing adjustments are necessary.
            - "skin_depth" (float): The calculated skin depth, which informs how the mesh should be adjusted to accurately model current variations.
        :rtype: Dict[str, Union[bool, float]]
        """
        if self.cacdm.mesh.matrix.rate_dependent_scaling_matrix and not self.cacdm.geometry.io_settings.load.load_from_yaml:
            matrix_mesh_options = self.cacdm.mesh.matrix

            filament_radius = self.geometry_class.filaments[0][0].rad

            mesh_size_matrix_inner = filament_radius * matrix_mesh_options.mesh_size_matrix_ratio_inner
            mesh_size_matrix_middle = filament_radius * matrix_mesh_options.mesh_size_matrix_ratio_middle
            mesh_size_matrix_outer = filament_radius * matrix_mesh_options.mesh_size_matrix_ratio_outer

            matrix = self.geometry_class.matrix

            rho_matrix = 1.81e-10 #self.cacdm.solve.material_properties.matrix.resistivity.constant
            mu0 = np.pi*4e-7
            skin_depth = np.sqrt( rho_matrix / (np.pi*mu0 * self.cacdm.solve.source_parameters.sine.frequency ) ) * self.cacdm.mesh.matrix.skindepth_scaling_factor
            # Function which approximates number of elements in a range given a linearly interpolated mesh size.
            number_of_elements_in_range = lambda a, b, x0, x1 : np.log(a*x1+b)/a - np.log(a*x0+b)/a if a != 0 else (x1-x0)/b
            
            # If the skindepth is smaller than the outer matrix, we only evaluate the outer matrix field to approximate number of elements in skindepth. 
            if skin_depth <= matrix[-1].rad -matrix[-2].rad: 
                a = (mesh_size_matrix_outer - mesh_size_matrix_middle) / (matrix[-1].rad - matrix[-2].rad)
                b = mesh_size_matrix_middle
                elements_in_skindepth = number_of_elements_in_range(a, b, matrix[-1].rad -matrix[-2].rad - skin_depth, matrix[-1].rad-matrix[-2].rad)

            elif skin_depth > matrix[-1].rad -matrix[-2].rad and skin_depth < matrix[-1].rad:
                # If the skindepth is larger than the outer matrix, we evaluate the inner and outer matrix field to approximate number of elements in skindepth. 
                a1 = (mesh_size_matrix_middle - mesh_size_matrix_inner) / (matrix[-2].rad)
                b1 = mesh_size_matrix_middle
                a2 = (mesh_size_matrix_outer - mesh_size_matrix_middle) / (matrix[-1].rad - matrix[-2].rad)
                b2 = mesh_size_matrix_middle
                elements_in_skindepth = number_of_elements_in_range(a1, b1, matrix[-1].rad-skin_depth, matrix[-2].rad) + number_of_elements_in_range(a2, b2, 0, matrix[-1].rad-matrix[-2].rad)

            else: # If the skindepth is greater than the matrix radius we do not want to act, and set the number of elements in skindepth to a high number.
                elements_in_skindepth = 1e10

            desired_elements_in_skindepth = self.cacdm.mesh.matrix.desired_elements_in_skindepth
            
            if elements_in_skindepth < desired_elements_in_skindepth:
                logger.info(f"The skin depth of the matrix is expected to contain only {round(elements_in_skindepth, 3)} elements. The mesh in the matrix has been adjusted to ensure at least {desired_elements_in_skindepth} elements in the skindepth.")
                return dict(frequency_dependent_meshing = True, skin_depth = skin_depth)

        return dict(frequency_dependent_meshing = False, skin_depth = 1e10)

    def matrix_field(self):
        """
        Creates a mesh size field for the matrix as a combination of three fields:
        - A distance-based interpolation between the matrix center and partition boundaries. The mesh size is linearly interpolated from the center to the boundary of the second matrix partition, and then to the outer matrix boundary.
        - Adjustments near the filaments to ensure a smooth transition from the filament mesh to the matrix mesh.
        - Frequency-dependent adjustments to capture skin depth effects.

        :return: The tag of the composite mesh size field created for the matrix.
        :rtype: int
        """
        matrix_mesh_options = self.cacdm.mesh.matrix
        # Choose the largest distance from the center of mass of the filaments to the filament boundary as the filament radius.
        # This approximation is done to account for the fact that the filaments may not be circular, but can have any shape.
        filament_rad = 0 
        for layer in self.geometry_class.filaments:
            for filament in layer:
                center_of_mass = gmsh.model.occ.get_center_of_mass(2, filament.surface_tag)
                for curve in filament.boundary_curves:
                    r = max([np.linalg.norm(np.array(center_of_mass) - curve.P1.pos), np.linalg.norm(np.array(center_of_mass) - curve.P2.pos)])
                    if r > filament_rad:
                        filament_rad = r

        mesh_size_matrix_inner = filament_rad * matrix_mesh_options.mesh_size_matrix_ratio_inner
        mesh_size_matrix_middle = filament_rad * matrix_mesh_options.mesh_size_matrix_ratio_middle
        mesh_size_matrix_outer = filament_rad * matrix_mesh_options.mesh_size_matrix_ratio_outer

        filament_edge_points = []
        for layer in self.geometry_class.filaments:
            for filament in layer:
                for curve in filament.boundary_curves:
                    filament_edge_points.append(curve.P1)
                    filament_edge_points.append(curve.P2)
        filament_outer_point = max(filament_edge_points, key=lambda p: np.linalg.norm(p.pos))
        # Distance from the outermost filament edge to the center of the matrix
        # This is done instead of referencing the middle matrix radius, to be able to mesh arbitrary geometries.
        filament_outer_point_dist = np.linalg.norm(filament_outer_point.pos) 

        matrix = self.geometry_class.matrix
        matrix_outer_point_dist = max([np.linalg.norm(p.pos) for curve in matrix[-1].boundary_curves for p in [curve.P1, curve.P2]]) # Distance from the outermost matrix edge to the center of the matrix

        matrix_mesh_size_fields = []

        # 'Distance from matrix center'- field:
        dist_from_center_field = gmsh.model.mesh.field.add("MathEval")
        gmsh.model.mesh.field.setString(dist_from_center_field, "F", "sqrt(x^2 + y^2)")

        # Mesh size-field in the middle partition of the matrix. The mesh size is large at the center and decreases linearily towards the edge of the middle matrix.
        interpField_innerToMiddle = gmsh.model.mesh.field.add("Threshold") 
        gmsh.model.mesh.field.setNumber(interpField_innerToMiddle, "InField", dist_from_center_field)
        gmsh.model.mesh.field.setNumber(interpField_innerToMiddle, "SizeMin", mesh_size_matrix_inner)
        gmsh.model.mesh.field.setNumber(interpField_innerToMiddle, "SizeMax", mesh_size_matrix_middle)
        gmsh.model.mesh.field.setNumber(interpField_innerToMiddle, "DistMin", 0)
        gmsh.model.mesh.field.setNumber(interpField_innerToMiddle, "DistMax", filament_outer_point_dist)

        # Restrict the middle field to the inner and middle matrix (only middle matrix in the case where we have a center filament and thus no inner matrix).
        middle_matrix_field = gmsh.model.mesh.field.add("Restrict")
        gmsh.model.mesh.field.setNumber(middle_matrix_field, "InField", interpField_innerToMiddle)
        gmsh.model.mesh.field.setNumbers(middle_matrix_field, "SurfacesList", [partition.surface_tag for partition in matrix[:-1]])
        matrix_mesh_size_fields.append(middle_matrix_field)
        
        # Mesh size-field in the outer partition of the matrix. The mesh size is small at the edge of the middle matrix and increases towards the outer edge.
        interpField_middleToOuter = gmsh.model.mesh.field.add("Threshold") 
        gmsh.model.mesh.field.setNumber(interpField_middleToOuter, "InField", dist_from_center_field)
        gmsh.model.mesh.field.setNumber(interpField_middleToOuter, "SizeMin", mesh_size_matrix_middle)
        gmsh.model.mesh.field.setNumber(interpField_middleToOuter, "SizeMax", mesh_size_matrix_outer)
        gmsh.model.mesh.field.setNumber(interpField_middleToOuter, "DistMin", filament_outer_point_dist)
        gmsh.model.mesh.field.setNumber(interpField_middleToOuter, "DistMax", matrix_outer_point_dist)

        # Restric the outer field to the outer matrix partition:
        outer_matrix_field = gmsh.model.mesh.field.add("Restrict")
        gmsh.model.mesh.field.setNumber(outer_matrix_field, "InField", interpField_middleToOuter)
        gmsh.model.mesh.field.setNumbers(outer_matrix_field, "SurfacesList", [matrix[-1].surface_tag])
        matrix_mesh_size_fields.append(outer_matrix_field)


        # The last field adjusts the matrix mesh size with respect to the distance from the filaments. The reason is that the filaments typically has a finer mesh than the matrix and 
        # the large mesh size gradient at the boundary between the matrix and the filaments can lead to a bad quality mesh. We thus want to gradually refine the mesh in the matrix close to the filaments. 
        # This is done by interpolating the mesh size based on the distance to the filament boundaries.
        meshSize_by_filament_field = self.matrix_threshold_to_filament_field([filament for layer in self.geometry_class.filaments for filament in layer])
        matrix_mesh_size_fields.append(meshSize_by_filament_field)

        frequency_dependent_meshing, skin_depth = self.evaluate_frequency_dependent_meshing().values()
        
        if frequency_dependent_meshing:
            # If the skindepth only consists of a few elements, we add an additional mesh size field in the matrix. This field is large outside of the skindepth, but
            # in the skindepth it is defined as the constant field which gives the skindepth the 'desired number of elements'. In between the two regions we interpolate between the two values.
            # logger.info(f"The skin depth of the matrix is expected to contain only {round(elements_in_skindepth, 3)} elements. The mesh in the matrix has been adjusted to ensure at least {desired_elements_in_skindepth} elements in the skindepth.")

            circular_field = gmsh.model.mesh.field.add("Ball")
            gmsh.model.mesh.field.setNumber(circular_field, "Radius", (matrix[-1].rad-skin_depth)*0.85 )
            # gmsh.model.mesh.field.setNumber(circular_field, "Thickness", (matrix[-1].rad-skin_depth)*0.15)
            gmsh.model.mesh.field.setNumber(circular_field, "VIn", mesh_size_matrix_inner)
            gmsh.model.mesh.field.setNumber(circular_field, "VOut", (skin_depth)/self.cacdm.mesh.matrix.desired_elements_in_skindepth )



            # -- Restrict the mesh size field to the matrix and the filaments not in the skindepth region -- #
            filaments_not_in_skindepth = [filament for layer in self.geometry_class.filaments for filament in layer if np.linalg.norm(filament.center_point.pos) < self.geometry_class.matrix[-1].rad - 2*skin_depth]
            
            circular_field_restricted = gmsh.model.mesh.field.add("Restrict")
            gmsh.model.mesh.field.setNumber(circular_field_restricted, "InField", circular_field)
            gmsh.model.mesh.field.setNumbers(circular_field_restricted, "SurfacesList", [partition.surface_tag for partition in matrix] + [filament.surface_tag for filament in filaments_not_in_skindepth] )
            # -- -- #

            self.cacdm.mesh.matrix.mesh_size_matrix_ratio_outer = skin_depth/self.cacdm.mesh.matrix.desired_elements_in_skindepth * 1/filament_rad

            filaments_in_skindepth = [filament for layer in self.geometry_class.filaments for filament in layer if np.linalg.norm(filament.center_point.pos) > self.geometry_class.matrix[-1].rad - 2*skin_depth]
            meshSize_by_filament_field = self.matrix_threshold_to_filament_field(filaments_in_skindepth) # Redefine the interpolation field to apply only to the filaments in the skindepth region.

            matrix_meshSize_field = gmsh.model.mesh.field.add("Min") 
            gmsh.model.mesh.field.setNumbers(matrix_meshSize_field, "FieldsList", [circular_field_restricted, meshSize_by_filament_field])
            return matrix_meshSize_field


        

        # Combine all the matrix mesh size fields by selecting the minimum value of all fields.
        matrix_meshSize_field = gmsh.model.mesh.field.add("Min") 
        gmsh.model.mesh.field.setNumbers(matrix_meshSize_field, "FieldsList", matrix_mesh_size_fields)

        return matrix_meshSize_field
    
    def air_field(self):
        """
        Generates a mesh size field for the air domain surrounding the model.
        The field linearly interpolates between the mesh size at the outer boundary of the matrix and the air boundary.

        :return: The tag of the created mesh size field for the air domain.
        :rtype: int
        """
        # Choose the largest distance from the center of mass of the filaments to the filament boundary as the filament radius.
        # This approximation is done to account for the fact that the filaments may not be circular, but can have any shape.
        filament_rad = 0 
        for layer in self.geometry_class.filaments:
            for filament in layer:
                center_of_mass = gmsh.model.occ.get_center_of_mass(2, filament.surface_tag)
                for curve in filament.boundary_curves:
                    r = max([np.linalg.norm(np.array(center_of_mass) - curve.P1.pos), np.linalg.norm(np.array(center_of_mass) - curve.P2.pos)])
                    if r > filament_rad:
                        filament_rad = r

        mesh_size_outer_matrix = filament_rad * self.cacdm.mesh.matrix.mesh_size_matrix_ratio_outer
        mesh_size_air_boundary = filament_rad * self.cacdm.mesh.air.max_mesh_size_ratio

        if self.geometry_class.air_composition:
            air_outer_boundary_curves = self.geometry_class.air_composition.boundary_curves
            air_inner_boundary_curves = sum(self.geometry_class.air_composition.inner_boundary_curves, [])
        else:
            air_outer_boundary_curves = self.geometry_class.air[0].boundary_curves
            air_inner_boundary_curves = sum(self.geometry_class.air[0].inner_boundary_curves, [])

        dist_from_outer_air_boundary_field = gmsh.model.mesh.field.add("Distance")
        gmsh.model.mesh.field.setNumbers(dist_from_outer_air_boundary_field, "CurvesList", [curve.tag for curve in air_outer_boundary_curves])

        dist_from_matrix_boundary = gmsh.model.mesh.field.add("Distance")
        gmsh.model.mesh.field.setNumbers(dist_from_matrix_boundary, "CurvesList", [curve.tag for curve in air_inner_boundary_curves])

        air_field = gmsh.model.mesh.field.add("MathEval")
        gmsh.model.mesh.field.setString(
            air_field, 
            "F", 
            f"( {mesh_size_air_boundary} * F{dist_from_matrix_boundary} + {mesh_size_outer_matrix} * F{dist_from_outer_air_boundary_field} ) / ( F{dist_from_matrix_boundary} + F{dist_from_outer_air_boundary_field} )"
            )
        return air_field

    def load_geometry_class(self, geom_folder):
        """
        Loads the TwistedStrand geometry class from a .pkl file saved within the specified geometry folder.
        The geometry class contains all the information about the strand geometry, facilitating mesh generation.

        :param geom_folder: The path to the folder where the geometry class .pkl file is saved.
        :type geom_folder: str

        :return: An instance of the TwistedStrand geometry class.
        :rtype: object
        """
        geom_save_file = os.path.join(geom_folder, f'{self.magnet_name}.pkl')

        with open(geom_save_file, "rb") as geom_save_file: # Unnecessary to return geom instead of setting self.geometry_class
            geom = pickle.load(geom_save_file)
        return geom

    def generate_mesh(self, geom_folder):
        """
        Generates the mesh for the entire model geometry based on combined mesh size fields.

        The method generates mesh size fields for the filament, matrix, and air domains separately and combines these fields to define a unified mesh size field for the full geometry. 
        The combined field is applied as the background mesh size field.

        :param geom_folder: The path to the folder containing the saved geometry class (.pkl file). This folder is used to load the geometry details required for mesh generation.
        :type geom_folder: str
        """

        # So far: Adds physical groups and generates an automatic mesh. To be further developed.
        self.geometry_class : TwistedStrand = self.load_geometry_class(geom_folder)
        self.geometry_class.update_tags()
        self.geometry_class.add_physical_groups()

        total_filament_boundary_distance_field = self.filament_field()
        matrixField = self.matrix_field()
        airField = self.air_field()

        total_meshSize_field = gmsh.model.mesh.field.add("Min")
        gmsh.model.mesh.field.setNumbers(total_meshSize_field, "FieldsList", [total_filament_boundary_distance_field, matrixField, airField])
    
        gmsh.model.mesh.field.setAsBackgroundMesh(total_meshSize_field)

        gmsh.option.setNumber("Mesh.MeshSizeFactor", self.cacdm.mesh.scaling_global)
        gmsh.option.setNumber("Mesh.MeshSizeExtendFromBoundary", 0)
        gmsh.option.setNumber("Mesh.MeshSizeFromPoints", 0)
        gmsh.option.setNumber("Mesh.MeshSizeFromCurvature", 0)

        # enforce center nodes
        if self.fdm.magnet.mesh.matrix.force_center_symmetry:
            dist = self.geometry_class.filaments[0][0].center_point.pos[0]*5e-2 # 5% of inner filament distance seems sufficient
            center = self.geometry_class.matrix[0].center_point.pos
            d_tag = gmsh.model.occ.add_point(center[0],center[1]-dist, center[2])
            u_tag = gmsh.model.occ.add_point(center[0],center[1]+dist, center[2])
            l_tag = gmsh.model.occ.add_point(center[0]-dist,center[1], center[2])
            r_tag = gmsh.model.occ.add_point(center[0]+dist,center[1], center[2])
            outDimTags, _ = gmsh.model.occ.fragment([(2, self.geometry_class.matrix[0].surface_tag)],[(0, gmsh.model.occ.add_point(center[0],center[1],center[2])),(0,d_tag),(0,u_tag),(0,l_tag),(0,r_tag)])  # 
            gmsh.model.occ.synchronize()
            # update the physical surface
            gmsh.model.removePhysicalGroups([(2, self.geometry_class.matrix[0].physical_surface_tag)])
            self.geometry_class.matrix[0].physical_surface_tag = gmsh.model.add_physical_group(2, [outDimTags[0][1]], tag=self.geometry_class.matrix[0].physical_surface_tag, name='Surface: Matrix partition 0')

        gmsh.model.mesh.generate(2)

    def generate_cuts(self):
        """
            Computes the cuts for imposing global quantities.
        """
        # Computes cuts, as needed for imposing global quantities
        if self.verbose:
            logger.info('(Co)homology computation')

        # a) Cut for the total current intensity (through OmegaCC)
        if self.fdm.magnet.geometry.type == 'periodic_square': 
            # only enforce periodic mesh - cut already defined in composite geometry
            self.set_periodic_mesh()
        else: 
            # cohomology cut for single air domain
            air_region = self.geometry_class.air[0]
            gmsh.model.mesh.addHomologyRequest("Homology", domainTags=[air_region.physical_inner_boundary_tags[0]], dims=[1])
            if self.fdm.magnet.geometry.type == 'strand_only':  
                gmsh.model.mesh.addHomologyRequest("Cohomology", domainTags=[air_region.physical_surface_tag], dims=[1])
            elif self.fdm.magnet.geometry.type == 'coil':
                # add semicircle arc as sudomain to force homology cut onto semicircle diameter
                gmsh.model.mesh.addHomologyRequest("Cohomology", domainTags=[air_region.physical_surface_tag], subdomainTags=[air_region.physical_cohomology_subdomain], dims=[1])

            cuts = gmsh.model.mesh.computeHomology()
            gmsh.model.mesh.clearHomologyRequests()
            gmsh.plugin.setString("HomologyPostProcessing", "PhysicalGroupsOfOperatedChains", str(cuts[1][1]))
            gmsh.plugin.setString("HomologyPostProcessing", "PhysicalGroupsOfOperatedChains2", str(cuts[0][1]))
            gmsh.plugin.run("HomologyPostProcessing")
            air_region.strand_bnd_cut_tag = cuts[1][1]+1 # The cut tag for the strand boundary	       

        # b) Cuts for the individual filaments
        for layer in self.geometry_class.filaments:
            for filament in layer:
                filBnd = filament.physical_boundary_tag
                gmsh.model.mesh.addHomologyRequest("Cohomology", domainTags=[filBnd],dims=[1])
        cuts = gmsh.model.mesh.computeHomology()

        filaments = sum(self.geometry_class.filaments, [])
        for i in range(len(cuts)):
            filaments[i].cut_tag = cuts[i][1]

    def set_periodic_mesh(self):
        """
        Enforces symmetric air mesh by mirroring the node positions from bottom to top and left to right.
        """
        air = self.geometry_class.air
        # link left and right
        dx, dy, dz = air[0].cut_curves[1].P2.pos - air[2].cut_curves[0].P2.pos        
        gmsh.model.mesh.set_periodic(1, [self.geometry_class.air_composition.physical_left_boundary_tag], [self.geometry_class.air_composition.physical_right_boundary_tag], [1,0,0, dx, 0,1,0, dy, 0,0,1, dz, 0,0,0,1])
        # link bottom and top
        dx, dy, dz = air[0].cut_curves[0].P2.pos - air[3].cut_curves[1].P2.pos
        gmsh.model.mesh.set_periodic(1, [self.geometry_class.air_composition.physical_bottom_boundary_tag], [self.geometry_class.air_composition.physical_top_boundary_tag], [1,0,0, dx, 0,1,0, dy, 0,0,1, dz, 0,0,0,1])

    def generate_regions_file(self):
        """
        Generates a .regions file for the GetDP solver, containing all necessary information about the model.
        The regions model contains data about physical surfaces, boundaries, points and cuts and is stored in the mesh folder.
        """
        regions_model = RegionsModel()
        
        """ -- Initialize data -- """
        regions_model.powered["Filaments"] = RegionsModelFiQuS.Powered()
        regions_model.powered["Filaments"].vol.numbers = [] # Filament physical surfaces tags
        regions_model.powered["Filaments"].vol.names = [] # Filament physical surfaces names
        regions_model.powered["Filaments"].surf.numbers = [] # Filament physical boundaries tags
        regions_model.powered["Filaments"].surf.names = [] # Filament physical boundaries names
        regions_model.powered["Filaments"].cochain.numbers = [] # Filament boundary cut tags
        regions_model.powered["Filaments"].curve.names = [] # Stores physical points at filament boundary (to fix phi=0)
        regions_model.powered["Filaments"].curve.numbers = [] # Stores physical points at filament boundary (to fix phi=0)

        regions_model.powered["Filaments"].surf_in.numbers = [] # Filament holes physical tags
        regions_model.powered["Filaments"].surf_in.names = [] # Filament holes physical names

        regions_model.insulator.curve.numbers = []  # Filament holes curves physical tags
        regions_model.insulator.curve.names = []    # Filament holes curves physical names

        regions_model.induced["Matrix"] = RegionsModelFiQuS.Induced()
        regions_model.induced["Matrix"].vol.numbers = [] # Matrix partition physical surfaces tags
        regions_model.induced["Matrix"].vol.names = [] # Matrix partition physical surfaces names
        regions_model.induced["Matrix"].surf_in.numbers = [] # Matrix partition physical boundaries tags
        regions_model.induced["Matrix"].surf_in.names = [] # Matrix partition physical boundaries names
        regions_model.induced["Matrix"].surf_out.numbers = [] # Strand physical outer boundary tag
        regions_model.induced["Matrix"].surf_out.names = [] # Strand physical outer boundary name
        regions_model.induced["Matrix"].cochain.numbers = [] # Strand outer boundary cut tag
        regions_model.induced["Matrix"].cochain.names = [] # Strand outer boundary cut name
        """ -- -- """

        for layer_i, layer in enumerate(self.geometry_class.filaments):
            for filament_j, filament in enumerate(layer):
                regions_model.powered["Filaments"].vol.numbers.append(filament.physical_surface_tag) # Surfaces in powered.vol
                if self.cacdm.geometry.io_settings.load.load_from_yaml:
                    regions_model.powered["Filaments"].vol.names.append(filament.material)
                else:
                    regions_model.powered["Filaments"].vol.names.append(filament.physical_surface_name)

                regions_model.powered["Filaments"].surf.numbers.append(filament.physical_boundary_tag) # Boundaries in powered.surf
                regions_model.powered["Filaments"].surf.names.append(filament.physical_boundary_name)

                regions_model.powered["Filaments"].cochain.numbers.append(filament.cut_tag)

                # Add physical point at filament boundary to fix phi=0
                regions_model.powered["Filaments"].curve.names.append(f"EdgePoint: filament_{layer_i+1}_{filament_j+1}")
                regions_model.powered["Filaments"].curve.numbers.append(filament.physicalEdgePointTag)

        for layer in self.geometry_class.filament_holes:
            for hole in layer:
                regions_model.powered["Filaments"].surf_in.numbers.append(hole.physical_surface_tag)
                regions_model.insulator.curve.numbers.append(hole.physical_boundary_tag)
                if self.cacdm.geometry.io_settings.load.load_from_yaml:
                    regions_model.powered["Filaments"].surf_in.names.append(hole.material)
                    regions_model.insulator.curve.names.append(hole.material)
                else:
                    regions_model.powered["Filaments"].surf_in.names.append(hole.physical_surface_name)
                    regions_model.insulator.curve.names.append(hole.physical_boundary_name)
                # Add physical point at hole boundary (inner filament boundary) to fix phi=0
                regions_model.powered["Filaments"].curve.numbers.append(hole.physicalEdgePointTag)

        for matrixPartition in self.geometry_class.matrix:
            regions_model.induced["Matrix"].vol.numbers.append(matrixPartition.physical_surface_tag)
            if self.cacdm.geometry.io_settings.load.load_from_yaml:
                regions_model.induced["Matrix"].vol.names.append(matrixPartition.material)
            else:
                regions_model.induced["Matrix"].vol.names.append(matrixPartition.physical_surface_name)

            regions_model.induced["Matrix"].surf_in.numbers.append(matrixPartition.physical_boundary_tag)
            regions_model.induced["Matrix"].surf_in.names.append(matrixPartition.physical_boundary_name)
        
        # Add manual cut regions depending on the selected geometry type
        if self.fdm.magnet.geometry.type == 'periodic_square':
            regions_model.air.cochain.numbers = self.geometry_class.air_composition.physical_cuts
            regions_model.air.cochain.names = ["Air vertical cut tag", "vertical cut strand boundary", "Air horizontal cut tag", "horizontal cut strand boundary"]
            air_region = self.geometry_class.air_composition

            regions_model.induced["Domain"] = RegionsModelFiQuS.Induced()
            regions_model.induced["Domain"].surf_out.numbers = [air_region.physical_top_boundary_tag, air_region.physical_bottom_boundary_tag, air_region.physical_left_boundary_tag, air_region.physical_right_boundary_tag]
            regions_model.induced["Domain"].surf_out.names = ["Domain boundary top", "Domain boundary bottom", "Domain boundary left", "Domain boundary right"]
        elif self.fdm.magnet.geometry.type == 'strand_only' or self.fdm.magnet.geometry.type == 'coil':
            regions_model.induced["Matrix"].cochain.numbers.append(self.geometry_class.air[0].strand_bnd_cut_tag)
            regions_model.induced["Matrix"].cochain.names.append("Strand boundary cut tag")
            air_region = self.geometry_class.air[0]

        regions_model.induced["Matrix"].surf_out.numbers.append(air_region.physical_inner_boundary_tags[0])
        regions_model.induced["Matrix"].surf_out.names.append(air_region.physical_inner_boundary_names[0])

        regions_model.air.vol.number = air_region.physical_surface_tag
        regions_model.air.vol.name = air_region.physical_surface_name

        regions_model.air.surf.number = air_region.physical_boundary_tag
        regions_model.air.surf.name = air_region.physical_boundary_name

        # Add physical gauge point in air to fix phi=0
        regions_model.air.point.names = ["Point at air boundary"]
        regions_model.air.point.numbers = [air_region.strand_bnd_physicalEdgePointTag]

        # HTCondor hack - store pickle file in geometry and mesh folder
        self.geometry_class.save(os.path.join(self.mesh_folder, f'{self.magnet_name}.pkl')) # Save the geometry-class to a pickle file
        FilesAndFolders.write_data_to_yaml(self.regions_file, regions_model.model_dump())


from typing import Literal, Optional

from pydantic import BaseModel, Field


class Condor(BaseModel):
    """
    Class for running using HTCondor queuing (only on Linux)
    """

    error: Optional[str] = Field(
        default=None,
        title="error",
        description="Error file name and extension, like error.txt",
    )
    output: Optional[str] = Field(
        default=None,
        title="output",
        description="Output file name and extension, like output.txt",
    )
    log: Optional[str] = Field(
        default=None,
        title="log",
        description="Log file name and extension, like log.txt",
    )
    request_cpus: Optional[int] = Field(
        default=None,
        title="request_cpus",
        description="Number of CPUs to request on each machine",
    )
    request_memory: Optional[str] = Field(
        default=None,
        title="request_memory",
        description="Amount of memory to request on each machine as a string (e.g., '16G')",
    ) 
    request_disk: Optional[str] = Field(
        default=None,
        title="request_disk",
        description="Amount of disk space to request on each machine as a string (e.g., '16G')",
    )
    singularity_image_path: Optional[str] = Field(
        default=None,
        title="SingularityImagePath",
        description="Full path to Singularity image",
    )
    cerngetdp_version: Optional[str] = Field(
        default=None,
        title="CERNGetDP Version",
        description="Version of CERNGetDP to be used",
    )
    should_transfer_files: Literal["YES", "NO"] = Field(
        default="YES",
        title="should_transfer_files",
        description="Sets if files should be transferred",
    )
    max_run_time: Optional[int] = Field(
        default=None,
        title="MaxRuntime",
        description=(
            "Specifies maximum run time in seconds to request for the job to go into"
            " the queue"
        ),
    )
    eos_relative_output_path: Optional[str] = Field(
        default=None,
        title="eos_relative_output_path",
        description=(
            "This is relative path in the user eos folder. This path gets appended to"
            " the root path: root://eosuser.cern.ch//eos/user/u/username"
        ),
    )
    big_mem_job: Optional[bool] = Field(
        default=None,
        title="BigMemJob",
        description=(
            "If true a machine with 1TB of RAM and 24 cores is requested. Expect longer"
            " queuing time"
        ),
    )

class Subproc(BaseModel):
    """
    Class for running using subprocess calls (on Windows)
    """

    executable: Optional[str] = Field(
        default=None,
        title="executable",
        description="Executable or script to run, like run_fiqus.py",
    )
    full_output_path: Optional[str] = Field(
        default=None,
        title="full_output_path",
        description="A full path to the output folder",
    )


class DataSettings(BaseModel):
    """
    Configuration for HTCondor and
    """

    GetDP_path: Optional[str] = Field(
        default=None,
        title="GetDP_path",
        description=(
            "Full path to GetDP executable. This is only needed and used on Windows"
        ),
    )

    base_path_model_files: Optional[str] = Field(
        default=None,
        title="base_path_model_files",
        description=(
            "Path to the base model folder where model files are stored that are needed to run FiQuS."
            "This is only needed when the files are not in the same folder as the input yaml (e.g., on HTCondor)"
        )
    )

    htcondor: Condor = Condor()
    subproc: Subproc = Subproc()


from pydantic import BaseModel, Field, ConfigDict
from typing import Dict, List, Optional, Union


class Area(BaseModel):
    Material: Optional[str] = Field(
        None, description="Material of the area", examples=["Cu", "NbTi", "Nb3Sn"]
    )
    Boundary: List[int] = (
        []
    )  # List of curves that define the closed boundary of the area
    InnerBoundaries: List[List[int]] = (
        []
    )  # List of lists of curves that define the closed boundaries of the holes in the area
    BoundaryThickness: Optional[float] = Field(
        None,
        description="Boundary layer thickness in meters",
        examples=1e-6,
    )  # Thickness of the boundary
    BoundaryMaterial: Optional[str] = Field(
        None,
        description="Material of the boundary",
        examples=["steel", "Cu", "direct", "etc."],
    )
    Layer: Optional[int] = Field(
        None,
        description="Filaments in the strand-model must be assigned to a layer. A layer is a collection of all filaments with the same radius from the center.",
    )
    LayerIndex: Optional[int] = Field(
        None, description="Index of the filament in the layer."
    )

# ========== GEOMETRY YAML CLASSES ========== #
class Material(BaseModel):
    Type: Optional[str] = Field(
        None, description="Type of material", examples=["NbTi", "Nb3Sn", "Cu"]
    )
    RRR: Optional[float] = Field(None, description="Residual resistivity ratio")
    T_ref_RRR_high: Optional[float] = Field(
        None, description="High reference temperature for RRR"
    )
    T_ref_RRR_low: Optional[float] = Field(
        None, description="Low reference temperature for RRR"
    )
    
    Resistivity: Optional[float] = Field(None, description="Resistivity of the material")
    model_config = ConfigDict(frozen=True)


class Point(BaseModel):
    Coordinates: List[float] = []


class Curve(BaseModel):
    Type: str
    Points: List[int] = []

    Contact: Optional[str] = Field(
        None,
        description="If the curve is a contact layer between two surfaces this represents the contact type of strands",
        examples=["crossing", "parallel"],
    )
    Thickness: Optional[float] = Field(
        None, description="Thickness of the contact layer"
    )
    Material: Optional[str] = Field(
        None,
        description="Material of the contact layer",
        examples=["steel", "direct", "Cu"],
    )


class GeometryParameters(BaseModel):
    Points: Dict[int, Point] = {}
    Curves: Dict[int, Curve] = {}
    Areas: Dict[int, Area] = {}

class DataDiffusionBarriers(BaseModel):
    FilamentResistances: List[float] = []
    DiffusionBarrierAreas: List[float] = []

class DataGlobalDiffusionBarrier(BaseModel):
    RegionTag: Optional[int] = Field(None)
    InternalRegionTag: Optional[int] = Field(None)
    ContactResistivity: Optional[float] = Field(None)

class SolutionParameters(BaseModel):
    Materials: Dict[str, Material] = {}
    DiffusionBarriers: DataDiffusionBarriers = DataDiffusionBarriers()
    Surfaces_excluded_from_TI: List[int] = []
    GlobalDiffusionBarrier: DataGlobalDiffusionBarrier = DataGlobalDiffusionBarrier()

class Conductor(BaseModel):
    Geometry: GeometryParameters = GeometryParameters()
    Solution: SolutionParameters = SolutionParameters()





import math
import os

from fiqus.geom_generators.GeometryConductorAC_Rutherford import Geometry
from fiqus.mesh_generators.MeshConductorAC_Rutherford import CableMesh
from fiqus.getdp_runners.RunGetdpConductorAC_Rutherford import Solve
from fiqus.post_processors.PostProcessAC_Rutherford import PostProcess
from fiqus.plotters.PlotPythonConductorAC import PlotPythonConductorAC

from fiqus.mains.MainConductorAC_Strand import MainConductorAC_Strand

class MainConductorAC_Rutherford(MainConductorAC_Strand):

    def generate_geometry(self, gui=False):
        """
        Generates the cable geometry.
        """
        os.chdir(self.geom_folder)
        g = Geometry(fdm=self.fdm, inputs_folder_path=self.inputs_folder_path, verbose=self.verbose)
        g.generate_cable_geometry(gui)



    def mesh(self, gui: bool = False):
        """ 
        Generates the mesh for the cable geometry.
        """
        os.chdir(self.mesh_folder)

        m = CableMesh(fdm=self.fdm, verbose=self.verbose)
        m.generate_mesh(self.geom_folder)
        m.generate_cuts()
        m.generate_regions_file()
        m.save_mesh(gui)

        return {"test": 0}
    
    def solve_and_postprocess_getdp(self, gui: bool = False):
        """
        Assembles the .pro-file from the template, then runs the simulation and the post-processing steps using GetDP.
        """
        os.chdir(self.solution_folder)

        # Checks if the strand type in the conductor data model is 'Homogenized', other types are not allowed (e.g., round) to avoid confusion as inputs in them won't be used
        if self.fdm.conductors[self.fdm.magnet.solve.conductor_name].strand.type != 'Homogenized':
            raise Exception(f"The strand type must be 'Homogenized' in the conductors section of the input .yaml file, any other type is not allowed.")

        s = Solve(self.fdm, self.GetDP_path, self.geom_folder, self.mesh_folder, self.verbose)
        s.read_excitation(inputs_folder_path=self.inputs_folder_path)
        # s.get_solution_parameters_from_yaml(inputs_folder_path=self.inputs_folder_path)
        s.assemble_pro()
        s.run_getdp(solve = True, postOperation = True, gui = gui)
        s.cleanup()
    
    def post_process_getdp(self, gui: bool = False):
        """ 
        Runs the post-processing steps trough GetDP.
        """
        os.chdir(self.solution_folder)

        s = Solve(self.fdm, self.GetDP_path, self.geom_folder, self.mesh_folder, self.verbose)
        s.read_excitation(inputs_folder_path=self.inputs_folder_path)
        # s.get_solution_parameters_from_yaml(inputs_folder_path=self.inputs_folder_path)
        s.assemble_pro()
        s.run_getdp(solve = False, postOperation = True, gui = gui)

    def post_process_python(self, gui: bool = False):
        """ 
        Runs the post-processing steps in the python PostProccess class.
        """
        os.chdir(self.solution_folder)

        p = PostProcess(self.solution_folder)
        p.show()

        return {'test': 0}


import os, re, logging
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

logger = logging.getLogger('FiQuS')


class PostProcess():
    """Class for python post processing of HomogenizedConductor simulations"""
    def __init__(self, fdm, solution_folder_path):
        self.fdm = fdm
        self.solution_folder_path = solution_folder_path
        # try to create ROHF plots
        self.plot_density_lines()
        self.fluxhyst()
        self.power_loss()
    
    def load_standard_txt(self, rel_file_path, skiprows=0, transpose=True):
        """This function loads standards txt files within the solution folder to numpy arrays.

        :param rel_file_path: relative path with file extension within the solution folder
        """
        abs_file_path = os.path.join(self.solution_folder_path, rel_file_path)
        data = None
        if os.path.isfile(abs_file_path):
            data = np.loadtxt(abs_file_path, skiprows=skiprows).T if transpose else np.loadtxt(abs_file_path, skiprows=skiprows)
        return data
    
    def fluxhyst(self):
        """This function tries to load and display the flux hysteresis (ROHF activated) in the solution data."""

        fluxhyst = self.load_standard_txt('txt_files\\fluxhyst.txt')
        #flux_average = self.load_standard_txt('txt_files\\flux_density_avg.txt')
        #flux_strand = self.load_standard_txt('txt_files\\flux_density_strand_avg.txt')
        
        voltage_rohf = self.load_standard_txt('txt_files\\voltage_rohf.txt')
        It = self.load_standard_txt('txt_files\\It.txt', skiprows=1)
        
        if fluxhyst is not None and It is not None:
            plt.figure(figsize=(10,5))
            plt.subplot(121)
            plt.title('Flux hysteresis')
            plt.plot(It[1], fluxhyst[1])
            #plt.plot(It[1], fluxhyst[1] - mu_0/(4*pi) * It[1])
            plt.xlabel('Transport current (A)')
            plt.ylabel('Internal Flux (Tm)')
            #plt.legend()
            
            plt.subplot(122)
            plt.title('ROHF voltage (unit length)')
            if voltage_rohf is not None: plt.plot(voltage_rohf[0], voltage_rohf[1])
            plt.xlabel('Time (s)')
            plt.ylabel('Voltage (V)')
            #plt.legend()
        else:
            logger.error("POSTPROC: No flux hysteresis data")

    def plot_density_lines(self):

        fluxdens_line = self.load_standard_txt('txt_files\\flux_density_line.txt', transpose=False)
        currentdens_line = self.load_standard_txt('txt_files\\current_density_line.txt', transpose=False)
        It = self.load_standard_txt('txt_files\\It.txt')
        
        idx = 10

        if fluxdens_line is not None and It is not None:
            N_steps = len(It[0])
            time = fluxdens_line[idx][1]
            x_vals = np.zeros(200)
            fluxdens = np.zeros(200)
            for i in range(0,200):
                x_vals[i] = fluxdens_line[idx+i*N_steps][2]
                fluxdens[i] = fluxdens_line[idx+i*N_steps][5]

            plt.figure()
            plt.title('fluxdens at '+str(time) +'s')
            plt.plot(x_vals, fluxdens)
            plt.xlabel('Position on x-axis')
            plt.ylabel('Flux density')

        if currentdens_line is not None and It is not None:
            N_steps = len(It[0])
            time = currentdens_line[idx][1]
            x_vals = np.zeros(200)
            currentdens = np.zeros(200)
            for i in range(0,200):
                x_vals[i] = currentdens_line[idx+i*N_steps][2]
                currentdens[i] = currentdens_line[idx+i*N_steps][7]

            plt.figure()
            plt.title('Current denisty at '+str(time) +'s')
            plt.plot(x_vals, currentdens)
            plt.xlabel('Position on x-axis')
            plt.ylabel('Current density')

    def power_loss(self):
        """This function tries to load and display the power loss from the solution data. Currently only supports ROHF associated losses. """

        power = self.load_standard_txt('txt_files\\power.txt')

        if power is not None and len(power) == 5:
            plt.figure()
            plt.title('Instantaneous power losses')
            plt.plot(power[0], power[2], label=r'$P_{\mathrm{irr}}$')
            plt.plot(power[0], power[3], label=r'$P_{\mathrm{eddy}}$')
            plt.plot(power[0], power[4], label=r'$P_{\mathrm{PL}}$')
            plt.legend()
        else:
            logger.error("POSTPROC: no/wrong power data")

    def show(self):
        """Utility funtion which is called in the end of the python post proc to display all figures at the same time"""
        plt.show() if len(plt.get_fignums()) > 0 else logger.info('- No postproc figures -')



import logging
import os
import gmsh
import time

from fiqus.plotters.PlotPythonMultipole import PlotPythonMultipole
from fiqus.utils.Utils import GmshUtils
from fiqus.utils.Utils import FilesAndFolders as Util
from fiqus.data import DataFiQuS as dF
from fiqus.data.DataRoxieParser import FiQuSGeometry
from fiqus.geom_generators.GeometryMultipole import Geometry
from fiqus.mesh_generators.MeshMultipole import Mesh
from fiqus.getdp_runners.RunGetdpMultipole import RunGetdpMultipole
from fiqus.getdp_runners.RunGetdpMultipole import AssignNaming
from fiqus.post_processors.PostProcessMultipole import PostProcess


class MainMultipole:
    def __init__(self, fdm: dF.FDM = None, rgd_path: str = None, verbose: bool = None, inputs_folder_path = None):
        """
        Main class for working with simulations for multipole type magnets
        :param fdm: FiQuS data model
        :param rgd_path: ROXIE geometry data path
        :param verbose: if True, more info is printed in the console
        """
        self.fdm = fdm
        self.rgd = rgd_path
        self.verbose = verbose

        self.GetDP_path = None
        self.geom_folder = None
        self.mesh_folder = None
        self.solution_folder = None
        self.inputs_folder_path = inputs_folder_path

    def force_symmetry(self):
        fdm = self.fdm.__deepcopy__()
        fdm.magnet.geometry.electromagnetics.symmetry = 'x'
        return fdm

    def generate_geometry(self, gui: bool = False):
        geom = Util.read_data_from_yaml(self.rgd, FiQuSGeometry)
        fdm = self.force_symmetry() if 'solenoid' in geom.Roxie_Data.coil.coils[1].type else self.fdm  # todo: this should be handled by pydantic
        if self.fdm.magnet.geometry.plot_preview:
            plotter = PlotPythonMultipole(geom, self.fdm)
            plotter.plot_coil_wedges()
        gg = Geometry(data=fdm, geom=geom, geom_folder=self.geom_folder, verbose=self.verbose)
        gg.saveHalfTurnCornerPositions()

        geometry_settings = {'EM': fdm.magnet.geometry.electromagnetics, 'TH': self.fdm.magnet.geometry.thermal}
        geometry_type_list = []

        if geometry_settings['EM'].create: geometry_type_list.append('EM')
        if geometry_settings['TH'].create: geometry_type_list.append('TH')
        for geometry_type in geometry_type_list:
            gg.saveStrandPositions(geometry_type)
            if any(geometry_settings[geometry_type].areas):
                gg.constructIronGeometry(geometry_settings[geometry_type].symmetry if geometry_type == 'EM' else 'none', geometry_settings[geometry_type], run_type = geometry_type)
            gg.constructCoilGeometry(geometry_type)
            if geometry_settings[geometry_type].with_wedges:
                gg.constructWedgeGeometry(geometry_settings[geometry_type].use_TSA if geometry_type == 'TH' else False)
            gmsh.model.occ.synchronize()
            if geometry_type == 'TH':
                if geometry_settings[geometry_type].use_TSA:
                    gg.constructThinShells(geometry_settings[geometry_type].with_wedges)
                else:
                    gg.constructInsulationGeometry()
                if geometry_settings[geometry_type].use_TSA_new:
                    gg.constructAdditionalThinShells()

            gg.buildDomains(geometry_type, geometry_settings[geometry_type].symmetry if geometry_type == 'EM' else 'none')
            if geometry_type == 'EM':
                gg.fragment()
            if geometry_type =='TH' and 'poles' in geometry_settings[geometry_type].areas:
                # make sure geometry is connected
                gmsh.model.occ.removeAllDuplicates()
                gmsh.model.occ.synchronize()

            gg.saveBoundaryRepresentationFile(geometry_type)
            gg.loadBoundaryRepresentationFile(geometry_type)
            gg.updateTags(geometry_type, geometry_settings[geometry_type].symmetry if geometry_type == 'EM' else 'none')
            gg.saveAuxiliaryFile(geometry_type)
            gg.clear()
        gg.ending_step(gui)

    def load_geometry(self, gui: bool = False):
        pass
        # gu = GmshUtils(self.geom_folder, self.verbose)
        # gu.initialize(verbosity_Gmsh=self.fdm.run.verbosity_Gmsh)
        # model_file = os.path.join(self.geom_folder, self.fdm.general.magnet_name)
        # gmsh.option.setString(name='Geometry.OCCTargetUnit', value='M')  # set units to meters
        # gmsh.open(model_file + '_EM.brep')
        # gmsh.open(model_file + '_TH.brep')
        # if gui: gu.launch_interactive_GUI()

    def pre_process(self, gui: bool = False):
        pass

    def load_geometry_for_mesh(self, run_type):
        gu = GmshUtils(self.geom_folder, self.verbose)
        gu.initialize(verbosity_Gmsh=self.fdm.run.verbosity_Gmsh)
        model_file = os.path.join(self.geom_folder, self.fdm.general.magnet_name)
        gmsh.option.setString(name='Geometry.OCCTargetUnit', value='M')  # set units to meters
        gmsh.open(model_file + f'_{run_type}.brep')

    def mesh(self, gui: bool = False):
        def _create_physical_group_for_reference(self):
            """
                This code generates the reference models for the FALCOND_C
            """
            ### hardcoded, we need only one reference
            if self.fdm.general.magnet_name == 'TEST_MULTIPOLE_FALCOND_C_TSA_COLLAR_POLE':
                CUT_REFERENCE = True # self specify loop and surf

                L_col = [56, 57, 58, 35, 16, 49, 48, 47]
                l1 = gmsh.model.occ.addLine(746,48)
                l2 = gmsh.model.occ.addLine(41,691)
                L_inner = list(range(854, 754-1, -1))
                loop1 = gmsh.model.occ.addCurveLoop([l1] + L_col + [l2] + L_inner)

                R_col = [52, 53, 54, 26, 6, 45, 44, 43]
                r1 = gmsh.model.occ.addLine(38,902)
                r2 = gmsh.model.occ.addLine(847,45)
                R_inner = list(range(910, 1010+1))
                loop2 = gmsh.model.occ.addCurveLoop([r2]+ R_col + [r1] + R_inner)

                surf1 = gmsh.model.occ.addPlaneSurface([loop1])
                surf2 = gmsh.model.occ.addPlaneSurface([loop2])
                surf = [surf1, surf2]  # tag

            else: raise Exception("Reference meshing is not implemented for this magnet.")
            gmsh.model.occ.synchronize()

            #add to physical groups / domains -> write to aux file
            file = os.path.join(self.geom_folder, self.fdm.general.magnet_name + '_TH.aux')
            with open(file) as f:
                lines = f.readlines()
            updated_lines = []
            flag = 0
            for line in lines:
                updated_lines.append(line)
                if flag == 0 and 'groups_entities:' in line:
                    flag = 1
                if flag==1 and ('ref_mesh: {}' in line):
                    if type(surf) == list:
                        updated_lines[-1] = f'    ref_mesh:\n      {self.fdm.magnet.solve.thermal.insulation_TSA.between_collar.material}: {surf}\n'
                    else:
                        updated_lines[-1] = f'    ref_mesh:\n      {self.fdm.magnet.solve.thermal.insulation_TSA.between_collar.material}: [{surf}]\n'
                        #if not CUT_REFERENCE else f'    ref_mesh:\n      {self.fdm.magnet.solve.thermal.insulation_TSA.between_collar.material}: '+ str(surf) +'\n'
                    flag = -1

            with open(file, 'w') as f:
                f.writelines(updated_lines)
            logger = logging.getLogger('FiQuS')
            logger.warning("Overwrite the .aux file to include the new domain")

        mm = Mesh(data=self.fdm, mesh_folder=self.mesh_folder, verbose=self.verbose) ## same mesh object is used for both thermal and EM
        geom = Util.read_data_from_yaml(self.rgd, FiQuSGeometry)
        fdm = self.force_symmetry() if 'solenoid' in geom.Roxie_Data.coil.coils[1].type else self.fdm
        geometry_settings = {'EM': fdm.magnet.geometry.electromagnetics, 'TH': self.fdm.magnet.geometry.thermal}
        mesh_settings = {'EM': fdm.magnet.mesh.electromagnetics, 'TH': fdm.magnet.mesh.thermal}

        mesh_type_list = []
        if mesh_settings['EM'].create: mesh_type_list.append('EM')
        if mesh_settings['TH'].create: mesh_type_list.append('TH')
        for physics_solved in mesh_type_list:
            self.load_geometry_for_mesh(physics_solved)
            if physics_solved == 'TH' and mesh_settings['TH'].reference.enabled and 'collar' in geometry_settings['TH'].areas:
                if self.fdm.magnet.geometry.thermal.use_TSA_new or self.fdm.magnet.geometry.thermal.use_TSA:
                    raise Exception('Reference solution is not implemented for collar with TSA')
                if 'iron_yoke' in geometry_settings['TH'].areas:
                    raise Exception('Reference solution is intended (read: hardcoded) without iron yoke')
                _create_physical_group_for_reference(self)
            mm.loadAuxiliaryFile(physics_solved)
            if any(geometry_settings[physics_solved].areas):
                mm.getIronCurvesTags(physics_solved)
            mm.defineMesh(geometry_settings[physics_solved], mesh_settings[physics_solved], physics_solved)
            mm.createPhysicalGroups(geometry_settings[physics_solved])
            mm.updateAuxiliaryFile(physics_solved)
            if geometry_settings[physics_solved].model_dump().get('use_TSA', False):
                mm.rearrangeThinShellsData() # rearrange data for the pro file, technically optional
            mm.assignRegionsTags(geometry_settings[physics_solved], mesh_settings[physics_solved])
            mm.saveRegionFile(physics_solved)
            mm.setMeshOptions()
            mm.generateMesh()
            mm.checkMeshQuality()
            mm.saveMeshFile(physics_solved)
            if geometry_settings[physics_solved].model_dump().get('use_TSA', False):
                mm.saveClosestNeighboursList()
                if self.fdm.magnet.mesh.thermal.isothermal_conductors: mm.selectMeshNodes(elements='conductors')
                if self.fdm.magnet.geometry.thermal.with_wedges and self.fdm.magnet.mesh.thermal.isothermal_wedges: mm.selectMeshNodes(elements='wedges')
            if geometry_settings[physics_solved].model_dump().get('use_TSA_new', False):
                mm.saveClosestNeighboursList_new_TSA()
            mm.saveHalfTurnCornerPositions()
            mm.saveRegionCoordinateFile(physics_solved)
            mm.clear()
        mm.ending_step(gui)
        return mm.mesh_parameters

    def load_mesh(self, gui: bool = False):
        gu = GmshUtils(self.geom_folder, self.verbose)
        gu.initialize(verbosity_Gmsh=self.fdm.run.verbosity_Gmsh)
        gmsh.open(f"{os.path.join(self.mesh_folder, self.fdm.general.magnet_name)}.msh")
        if gui: gu.launch_interactive_GUI()

    def solve_and_postprocess_getdp(self, gui: bool = False):
        an = AssignNaming(data=self.fdm)
        rg = RunGetdpMultipole(data=an, solution_folder=self.solution_folder, GetDP_path=self.GetDP_path,
                               verbose=self.verbose)
        rg.loadRegionFiles()
        if self.fdm.magnet.solve.thermal.solve_type and self.fdm.magnet.geometry.thermal.use_TSA:
            rg.loadRegionCoordinateFile()
        rg.assemblePro()
        start_time = time.time()
        rg.solve_and_postprocess()
        rg.ending_step(gui)
        return time.time() - start_time

    def solve_and_postprocess_getdp(self, gui: bool = False):
        an = AssignNaming(data=self.fdm)
        rg = RunGetdpMultipole(data=an, solution_folder=self.solution_folder, GetDP_path=self.GetDP_path,
                               verbose=self.verbose)
        rg.loadRegionFiles()
        if self.fdm.magnet.solve.thermal.solve_type and self.fdm.magnet.geometry.thermal.use_TSA:
            rg.loadRegionCoordinateFile()
            rg.read_aux_file(os.path.join(self.mesh_folder, f"{self.fdm.general.magnet_name}_EM.aux"))
            rg.extract_half_turn_blocks()
        if self.fdm.magnet.geometry.thermal.use_TSA_new:
            rg.read_aux_file(os.path.join(self.mesh_folder,
                                          f"{self.fdm.general.magnet_name}_TH.aux"))  # now load the thermal aux file
            rg.extract_specific_TSA_lines()
        rg.assemblePro()
        start_time = time.time()
        rg.solve_and_postprocess()
        rg.ending_step(gui)
        return time.time() - start_time


    def post_process_getdp(self, gui: bool = False):
        an = AssignNaming(data=self.fdm)
        rg = RunGetdpMultipole(data=an, solution_folder=self.solution_folder, GetDP_path=self.GetDP_path, verbose=self.verbose)
        rg.loadRegionFiles()
        if self.fdm.magnet.solve.thermal.solve_type and self.fdm.magnet.geometry.thermal.use_TSA:
            rg.loadRegionCoordinateFile()
        rg.assemblePro()
        rg.postprocess()
        rg.ending_step(gui)

    def post_process_python(self, gui: bool = False):
        if self.fdm.run.type == 'post_process_python_only':
            an = AssignNaming(data=self.fdm)
            data = an.data
        else: data = self.fdm

        run_types = []
        if self.fdm.magnet.solve.electromagnetics.solve_type: run_types.append('EM')
        if self.fdm.magnet.solve.thermal.solve_type: run_types.append('TH')
        pp_settings = {'EM': self.fdm.magnet.postproc.electromagnetics, 'TH': self.fdm.magnet.postproc.thermal}
        pp = PostProcess(data=data, solution_folder=self.solution_folder, verbose=self.verbose)
        for run_type in run_types:
            pp.prepare_settings(pp_settings[run_type])
            pp.loadStrandPositions(run_type)
            pp.loadAuxiliaryFile(run_type)
            if pp_settings[run_type].plot_all != 'False': pp.loadHalfTurnCornerPositions()
            if pp_settings[run_type].model_dump().get('take_average_conductor_temperature', False): pp.loadRegionFile()
            pp.postProcess(pp_settings[run_type])
            if run_type == 'EM' and self.fdm.magnet.geometry.electromagnetics.symmetry != 'none': pp.completeMap2d()
            pp.clear()
        pp.ending_step(gui)
        return pp.postprocess_parameters

    def plot_python(self):
        os.chdir(self.solution_folder)
        p = PlotPythonMultipole(self.fdm, self.fdm)
        p.plot_coil_wedges()
